import numpy as np
import pandas as pd
import pytest

import spotify_confidence
from spotify_confidence.analysis.constants import (
    INCREASE_PREFFERED,
    DECREASE_PREFFERED,
    POINT_ESTIMATE,
    CI_LOWER,
    CI_UPPER,
    P_VALUE,
    ADJUSTED_LOWER,
    ADJUSTED_UPPER,
    DIFFERENCE,
    BONFERRONI,
    BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY,
    CORRECTION_METHODS,
    SPOT_1,
    CORRECTION_METHODS_THAT_SUPPORT_CI,
    POWERED_EFFECT,
    REQUIRED_SAMPLE_SIZE,
)


class TestPoweredEffectContinuousSingleMetric(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "control",
                    "test2",
                ],
                "nr_of_items": [
                    500,
                    8,
                    100,
                ],
                "nr_of_items_sumsq": [
                    2500,
                    12,
                    150,
                ],
                "users": [
                    1010,
                    22,
                    150,
                ],
                "metric_name": ["metricA", "metricA", "metricA"],
                "minimum_detectable_effect": [0.2, 0.2, 0.2],
            }
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns=["variation_name"],
            ordinal_group_column=None,
            interval_size=0.95,
            metric_column="metric_name",
            treatment_column="variation_name",
            power=0.8,
        )

    def test_powered_effect1(self):
        powered_effect = self.test.difference(
            level_1=("control", "metricA"),
            level_2=("test", "metricA"),
            minimum_detectable_effects_column="minimum_detectable_effect",
        )
        assert np.isclose(powered_effect[POWERED_EFFECT][0], 0.3972, atol=0.001)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][0], 30792, atol=100)

    def test_powered_effect2(self):
        powered_effect = self.test.difference(
            level_1=("control", "metricA"),
            level_2=("test2", "metricA"),
            minimum_detectable_effects_column="minimum_detectable_effect",
        )
        assert np.isclose(powered_effect[POWERED_EFFECT][0], 0.4208, atol=0.001)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][0], 5760, atol=100)


class TestPoweredEffectContinuousMultipleSuccessMetrics(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": ["test", "control", "test2", "test", "control", "test2"],
                "nr_of_items": [
                    500,
                    8,
                    100,
                    500,
                    8,
                    100,
                ],
                "nr_of_items_sumsq": [
                    2500,
                    12,
                    150,
                    2500,
                    12,
                    150,
                ],
                "users": [
                    1010,
                    22,
                    150,
                    1010,
                    22,
                    150,
                ],
                "metric_name": ["metricA", "metricA", "metricA", "metricB", "metricB", "metricB"],
                "minimum_detectable_effect": [0.2, 0.2, 0.2, 0.2, 0.2, 0.2],
            }
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns=["variation_name", "metric_name"],
            ordinal_group_column=None,
            interval_size=0.95,
            power=0.8,
        )

    def test_powered_effect1(self):
        powered_effect = self.test.multiple_difference(
            level="control",
            groupby="metric_name",
            level_as_reference=True,
            minimum_detectable_effects_column="minimum_detectable_effect",
        )
        assert np.isclose(powered_effect[POWERED_EFFECT][0], 0.4735, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][1], 0.5016, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][2], 0.4735, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][3], 0.5016, atol=0.001)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][0], 43746, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][1], 8183, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][2], 43746, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][3], 8183, atol=100)


class TestPoweredEffectContinuousMultipleMetricTypes(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": ["test", "control", "test2", "test", "control", "test2"],
                "nr_of_items": [
                    500,
                    8,
                    100,
                    500,
                    8,
                    100,
                ],
                "nr_of_items_sumsq": [
                    2500,
                    12,
                    150,
                    2500,
                    12,
                    150,
                ],
                "users": [
                    1010,
                    22,
                    150,
                    1010,
                    22,
                    150,
                ],
                "metric_name": ["metricA", "metricA", "metricA", "metricB", "metricB", "metricB"],
                "non_inferiority_margin": [None, None, None, 0.01, 0.01, 0.01],
                "preferred_direction": [None, None, None, "increase", "increase", "increase"],
            }
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns=["variation_name", "metric_name"],
            ordinal_group_column=None,
            interval_size=0.95,
            correction_method="spot-1-bonferroni",
            metric_column="metric_name",
            treatment_column="variation_name",
            power=0.8,
        )

    def test_powered_effect(self):
        powered_effect = self.test.multiple_difference(
            level="control", groupby="metric_name", level_as_reference=True, non_inferiority_margins=True
        )
        assert np.isclose(powered_effect[POWERED_EFFECT][0], 0.4995, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][1], 0.5291, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][2], 0.4596, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][3], 0.4869, atol=0.001)
        assert powered_effect[REQUIRED_SAMPLE_SIZE][0] is None
        assert powered_effect[REQUIRED_SAMPLE_SIZE][1] is None
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][2], 16487886, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][3], 3083846, atol=100)


class TestPoweredEffectContinuousMultipleMetricsSegments(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                ],
                "nr_of_items": [
                    500,
                    8,
                    100,
                    500,
                    8,
                    100,
                    500,
                    8,
                    100,
                    500,
                    8,
                    100,
                ],
                "nr_of_items_sumsq": [
                    2500,
                    12,
                    150,
                    2500,
                    12,
                    150,
                    2500,
                    12,
                    150,
                    2500,
                    12,
                    150,
                ],
                "users": [
                    1010,
                    22,
                    150,
                    1010,
                    22,
                    150,
                    1010,
                    22,
                    150,
                    1010,
                    22,
                    150,
                ],
                "metric_name": [
                    "metricA",
                    "metricA",
                    "metricA",
                    "metricB",
                    "metricB",
                    "metricB",
                    "metricA",
                    "metricA",
                    "metricA",
                    "metricB",
                    "metricB",
                    "metricB",
                ],
                "non_inferiority_margin": [None, None, None, 0.01, 0.01, 0.01, None, None, None, 0.01, 0.01, 0.01],
                "preferred_direction": [
                    None,
                    None,
                    None,
                    "increase",
                    "increase",
                    "increase",
                    None,
                    None,
                    None,
                    "increase",
                    "increase",
                    "increase",
                ],
                "segment": ["us", "us", "us", "us", "us", "us", "se", "se", "se", "se", "se", "se"],
            }
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns=["variation_name", "metric_name", "segment"],
            ordinal_group_column=None,
            interval_size=0.95,
            correction_method="spot-1-bonferroni",
            metric_column="metric_name",
            treatment_column="variation_name",
            power=0.8,
        )

    def test_powered_effect(self):
        powered_effect = self.test.multiple_difference(
            level="control", groupby=["metric_name", "segment"], level_as_reference=True, non_inferiority_margins=True
        )
        assert np.isclose(powered_effect[POWERED_EFFECT][0], 0.5358, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][1], 0.5676, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][2], 0.5358, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][3], 0.5676, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][4], 0.4995, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][5], 0.5291, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][6], 0.4995, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][7], 0.5291, atol=0.001)

        assert powered_effect[REQUIRED_SAMPLE_SIZE][0] is None
        assert powered_effect[REQUIRED_SAMPLE_SIZE][1] is None
        assert powered_effect[REQUIRED_SAMPLE_SIZE][2] is None
        assert powered_effect[REQUIRED_SAMPLE_SIZE][3] is None
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][4], 19475238, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][5], 3642591, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][6], 19475238, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][7], 3642591, atol=100)


class TestPoweredEffectContinuousMultipleMetricsSegments2(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                ],
                "nr_of_items": [
                    500,
                    8,
                    100,
                    500,
                    8,
                    100,
                    500,
                    8,
                    100,
                    500,
                    8,
                    100,
                ],
                "nr_of_items_sumsq": [
                    2500,
                    12,
                    150,
                    2500,
                    12,
                    150,
                    2500,
                    12,
                    150,
                    2500,
                    12,
                    150,
                ],
                "users": [
                    1010,
                    22,
                    150,
                    1010,
                    22,
                    150,
                    1010,
                    22,
                    150,
                    1010,
                    22,
                    150,
                ],
                "metric_name": [
                    "metricA",
                    "metricA",
                    "metricA",
                    "metricB",
                    "metricB",
                    "metricB",
                    "metricA",
                    "metricA",
                    "metricA",
                    "metricB",
                    "metricB",
                    "metricB",
                ],
                "non_inferiority_margin": [0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01],
                "preferred_direction": [
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                ],
                "segment": ["us", "us", "us", "us", "us", "us", "se", "se", "se", "se", "se", "se"],
            }
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns=["variation_name", "metric_name", "segment"],
            ordinal_group_column=None,
            interval_size=0.95,
            correction_method="spot-1-bonferroni",
            metric_column="metric_name",
            treatment_column="variation_name",
            power=0.8,
        )

    def test_powered_effect(self):
        powered_effect = self.test.multiple_difference(
            level="control", groupby=["metric_name", "segment"], level_as_reference=True, non_inferiority_margins=True
        )
        assert np.isclose(powered_effect[POWERED_EFFECT][0], 0.4995, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][1], 0.5291, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][2], 0.4995, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][3], 0.5291, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][4], 0.4995, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][5], 0.5291, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][6], 0.4995, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][7], 0.5291, atol=0.001)

        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][0], 19475238, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][1], 3642591, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][2], 19475238, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][3], 3642591, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][4], 19475238, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][5], 3642591, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][6], 19475238, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][7], 3642591, atol=100)


class TestPoweredEffectContinuousMultipleMetricsSegments3(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                ],
                "nr_of_items": [
                    500,
                    8,
                    100,
                    500,
                    8,
                    100,
                    500,
                    8,
                    100,
                    500,
                    8,
                    100,
                ],
                "nr_of_items_sumsq": [
                    2500,
                    12,
                    150,
                    2500,
                    12,
                    150,
                    2500,
                    12,
                    150,
                    2500,
                    12,
                    150,
                ],
                "users": [
                    1010,
                    22,
                    150,
                    1010,
                    22,
                    150,
                    1010,
                    22,
                    150,
                    1010,
                    22,
                    150,
                ],
                "metric_name": [
                    "metricA",
                    "metricA",
                    "metricA",
                    "metricB",
                    "metricB",
                    "metricB",
                    "metricA",
                    "metricA",
                    "metricA",
                    "metricB",
                    "metricB",
                    "metricB",
                ],
                "minimum_detectable_effect": [0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02, 0.02],
                "preferred_direction": [
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                ],
                "segment": ["us", "us", "us", "us", "us", "us", "se", "se", "se", "se", "se", "se"],
            }
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns=["variation_name", "metric_name", "segment"],
            ordinal_group_column=None,
            interval_size=0.95,
            correction_method="spot-1-bonferroni",
            metric_column="metric_name",
            treatment_column="variation_name",
            power=0.8,
        )

    def test_powered_effect(self):
        powered_effect = self.test.multiple_difference(
            level="control",
            groupby=["metric_name", "segment"],
            level_as_reference=True,
            minimum_detectable_effects_column="minimum_detectable_effect",
        )
        assert np.isclose(powered_effect[POWERED_EFFECT][0], 0.4735, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][1], 0.5016, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][2], 0.4735, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][3], 0.5016, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][4], 0.4735, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][5], 0.5016, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][6], 0.4735, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][7], 0.5016, atol=0.001)

        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][0], 4374482, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][1], 818191, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][2], 4374482, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][3], 818191, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][4], 4374482, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][5], 818191, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][6], 4374482, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][7], 818191, atol=100)


class TestPoweredEffectBinary(object):
    def setup(self):
        np.random.seed(123)

        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "test",
                    "control",
                    "control",
                    "test2",
                    "test2",
                    "test3",
                    "test3",
                    "test",
                    "test",
                    "control",
                    "control",
                    "test2",
                    "test2",
                    "test3",
                    "test3",
                ],
                "success": [50, 60, 140, 140, 10, 20, 20, 20, 50, 60, 140, 140, 10, 20, 20, 20],
                "total": [100, 100, 200, 200, 50, 50, 60, 60, 100, 100, 200, 200, 50, 50, 60, 60],
                "country": [
                    "us",
                    "ca",
                    "us",
                    "ca",
                    "us",
                    "ca",
                    "us",
                    "ca",
                    "us",
                    "ca",
                    "us",
                    "ca",
                    "us",
                    "ca",
                    "us",
                    "ca",
                ],
                "metric_name": [
                    "metricA",
                    "metricA",
                    "metricA",
                    "metricA",
                    "metricA",
                    "metricA",
                    "metricA",
                    "metricA",
                    "metricB",
                    "metricB",
                    "metricB",
                    "metricB",
                    "metricB",
                    "metricB",
                    "metricB",
                    "metricB",
                ],
                "preferred_direction": [
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                ],
                "non_inferiority_margin": [
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                ],
            }
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="success",
            numerator_sum_squares_column=None,
            denominator_column="total",
            categorical_group_columns=["country", "variation_name"],
            interval_size=0.95,
            correction_method="spot-1-bonferroni",
            metric_column="metric_name",
            treatment_column="variation_name",
            power=0.8,
        )

    def test_powered_effect(self):
        powered_effect = self.test.multiple_difference(
            level="control", groupby=["metric_name", "country"], level_as_reference=True, non_inferiority_margins=True
        )
        #  assert np.isclose(powered_effect[POWERED_EFFECT][0], 0.1984, atol=0.001)
        #  assert np.isclose(powered_effect[POWERED_EFFECT][1], 0.2599, atol=0.001)
        #  assert np.isclose(powered_effect[POWERED_EFFECT][2], 0.2411, atol=0.001)
        #  assert np.isclose(powered_effect[POWERED_EFFECT][3], 0.1984, atol=0.001)
        #  assert np.isclose(powered_effect[POWERED_EFFECT][4], 0.2599, atol=0.001)
        #  assert np.isclose(powered_effect[POWERED_EFFECT][5], 0.2411, atol=0.001)
        #  assert np.isclose(powered_effect[POWERED_EFFECT][6], 0.2062, atol=0.001)
        #  assert np.isclose(powered_effect[POWERED_EFFECT][7], 0.2663, atol=0.001)
        #  assert np.isclose(powered_effect[POWERED_EFFECT][8], 0.2479, atol=0.001)
        #  assert np.isclose(powered_effect[POWERED_EFFECT][9], 0.2062, atol=0.001)
        #  assert np.isclose(powered_effect[POWERED_EFFECT][10], 0.2663, atol=0.001)
        #  assert np.isclose(powered_effect[POWERED_EFFECT][11], 0.2479, atol=0.001)

        assert powered_effect[REQUIRED_SAMPLE_SIZE][0] is None
        assert powered_effect[REQUIRED_SAMPLE_SIZE][1] is None
        assert powered_effect[REQUIRED_SAMPLE_SIZE][2] is None
        assert powered_effect[REQUIRED_SAMPLE_SIZE][3] is None
        assert powered_effect[REQUIRED_SAMPLE_SIZE][4] is None
        assert powered_effect[REQUIRED_SAMPLE_SIZE][5] is None
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][6], 260541, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][7], 361863, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][8], 326159, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][9], 260541, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][10], 361863, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][11], 326159, atol=100)


class TestPoweredEffectBinaryOnlyGuardrail(object):
    def setup(self):
        np.random.seed(123)

        self.data = pd.DataFrame(
            {
                "variation_name": ["test", "test", "control", "control", "test2", "test2", "test3", "test3"],
                "success": [50, 60, 140, 140, 20, 20, 20, 20],
                "total": [100, 100, 200, 200, 50, 50, 60, 60],
                "country": ["us", "ca", "us", "ca", "us", "ca", "us", "ca"],
                "metric_name": [
                    "metricB",
                    "metricB",
                    "metricB",
                    "metricB",
                    "metricB",
                    "metricB",
                    "metricB",
                    "metricB",
                ],
                "preferred_direction": [
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                ],
                "non_inferiority_margin": [0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01],
            }
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="success",
            numerator_sum_squares_column=None,
            denominator_column="total",
            categorical_group_columns=["country", "variation_name"],
            interval_size=0.95,
            correction_method="spot-1-bonferroni",
            metric_column="metric_name",
            treatment_column="variation_name",
            power=0.8,
        )

    def test_powered_effect(self):
        powered_effect = self.test.multiple_difference(
            level="control", groupby=["metric_name", "country"], level_as_reference=True, non_inferiority_margins=True
        )
        assert np.isclose(powered_effect[POWERED_EFFECT][0], 0.1816, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][1], 0.2344, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][2], 0.2182, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][3], 0.1816, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][4], 0.2344, atol=0.001)
        assert np.isclose(powered_effect[POWERED_EFFECT][5], 0.2182, atol=0.001)

        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][0], 201905, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][1], 280423, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][2], 252755, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][3], 201905, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][4], 280423, atol=100)
        assert np.isclose(powered_effect[REQUIRED_SAMPLE_SIZE][5], 252755, atol=100)


class TestBinary(object):
    def setup(self):
        np.random.seed(123)

        self.data = pd.DataFrame(
            {
                "variation_name": ["test", "control", "test2", "test3"],
                "success": [50, 40, 10, 20],
                "total": [100, 100, 50, 60],
            }
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="success",
            numerator_sum_squares_column=None,
            denominator_column="total",
            categorical_group_columns="variation_name",
            correction_method="bonferroni",
        )

    def test_init_sumsq_sum(self):
        spotify_confidence.ZTest(
            self.data,
            numerator_column="success",
            numerator_sum_squares_column="success",
            denominator_column="total",
            categorical_group_columns=["variation_name"],
        )

    def test_summary(self):
        summary_df = self.test.summary()
        assert len(summary_df) == len(self.data)

    def test_difference(self):
        difference_df = self.test.difference(level_1="control", level_2="test", absolute=True)
        assert len(difference_df) == 1
        assert difference_df["difference"][0] == 0.5 - 0.4

    def test_difference_absolute_false(self):
        difference_df = self.test.difference(level_1="control", level_2="test", absolute=False)
        assert len(difference_df) == 1
        assert difference_df["difference"][0] == (0.5 - 0.4) / 0.4

    def test_multiple_difference(self):
        difference_df = self.test.multiple_difference(level="control", level_as_reference=True)
        assert len(difference_df) == self.data.variation_name.unique().size - 1
        assert difference_df["difference"][0] == 0.5 - 0.4

    def test_multiple_difference_level_as_reference_false(self):
        difference_df_true_true = self.test.multiple_difference(
            level="control", level_as_reference=True, absolute=True
        )

        difference_df = self.test.multiple_difference(level="control", level_as_reference=False, absolute=True)
        assert len(difference_df) == self.data.variation_name.unique().size - 1
        assert np.allclose(difference_df["difference"], -difference_df_true_true["difference"], atol=0)
        assert np.allclose(difference_df["ci_lower"], -difference_df_true_true["ci_upper"], atol=0)
        assert np.allclose(difference_df["ci_upper"], -difference_df_true_true["ci_lower"], atol=0)
        assert np.allclose(difference_df["p-value"], difference_df_true_true["p-value"], atol=0)

    def test_multiple_difference_absolute_false(self):
        control_mean = self.test.summary().query("variation_name == 'control'")["point_estimate"].values[0]
        difference_df_true_true = self.test.multiple_difference(
            level="control", level_as_reference=True, absolute=True
        )

        difference_df = self.test.multiple_difference(level="control", level_as_reference=True, absolute=False)
        assert len(difference_df) == self.data.variation_name.unique().size - 1
        assert np.allclose(difference_df["difference"], difference_df_true_true["difference"] / control_mean, atol=0)
        assert np.allclose(difference_df["ci_lower"], difference_df_true_true["ci_lower"] / control_mean, atol=0)
        assert np.allclose(difference_df["ci_upper"], difference_df_true_true["ci_upper"] / control_mean, atol=0)
        assert np.allclose(difference_df["p-value"], difference_df_true_true["p-value"], atol=0)

    def test_multiple_difference_level_as_reference_false_absolute_false(self):
        reference_mean = self.test.summary().query("variation_name != 'control'")["point_estimate"]
        difference_df_true_true = self.test.multiple_difference(
            level="control", level_as_reference=True, absolute=True
        )

        difference_df = self.test.multiple_difference(level="control", level_as_reference=False, absolute=False)
        assert len(difference_df) == self.data.variation_name.unique().size - 1
        assert np.allclose(
            difference_df["difference"], -difference_df_true_true["difference"] / reference_mean.values, atol=0
        )
        assert np.allclose(
            difference_df["ci_lower"], -difference_df_true_true["ci_upper"] / reference_mean.values, atol=0
        )
        assert np.allclose(
            difference_df["ci_upper"], -difference_df_true_true["ci_lower"] / reference_mean.values, atol=0
        )
        assert np.allclose(difference_df["p-value"], difference_df_true_true["p-value"], atol=0)

    def test_summary_plot(self):
        chart_grid = self.test.summary_plot()
        assert len(chart_grid.charts) == 1

    def test_difference_plot(self):
        chartgrid = self.test.difference_plot(level_1="control", level_2="test")
        assert len(chartgrid.charts) == 1

    def test_multiple_difference_plot(self):
        chartgrid = self.test.multiple_difference_plot(level="control", level_as_reference=True)
        assert len(chartgrid.charts) == 1


class TestCategoricalBinary(object):
    def setup(self):
        np.random.seed(123)

        self.data = pd.DataFrame(
            {
                "variation_name": ["test", "test", "control", "control", "test2", "test2", "test3", "test3"],
                "success": [50, 60, 40, 140, 10, 20, 20, 20],
                "total": [100, 100, 100, 200, 50, 50, 60, 60],
                "country": ["us", "ca", "us", "ca", "us", "ca", "us", "ca"],
            }
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="success",
            numerator_sum_squares_column=None,
            denominator_column="total",
            categorical_group_columns=["country", "variation_name"],
        )

    def test_init_sumsq_sum(self):
        spotify_confidence.ZTest(
            self.data,
            numerator_column="success",
            numerator_sum_squares_column="success",
            denominator_column="total",
            categorical_group_columns=["variation_name", "country"],
        )

    def test_init_sumsq_sum_one_country(self):
        spotify_confidence.ZTest(
            self.data.query('country == "us"'),
            numerator_column="success",
            numerator_sum_squares_column="success",
            denominator_column="total",
            categorical_group_columns="variation_name",
        )

    def test_summary(self):
        summary_df = self.test.summary()
        assert len(summary_df) == len(self.data)

    def test_difference(self):
        difference_df = self.test.difference(level_1=("us", "control"), level_2=("ca", "test"))
        assert len(difference_df) == 1

    def test_difference_groupby(self):
        difference_df = self.test.difference(level_1="control", level_2="test", groupby="country")
        assert len(difference_df) == self.data.country.unique().size

    def test_multiple_difference(self):
        difference_df = self.test.multiple_difference(level=("us", "control"), level_as_reference=True, verbose=True)
        assert len(difference_df) == (
            (self.data.variation_name.unique().size - 1) * self.data.country.unique().size
            + self.data.country.unique().size
            - 1
        )
        n_comp = len(difference_df)
        assert np.allclose(
            difference_df.apply(lambda row: min(row[P_VALUE] * n_comp, 1), axis=1),
            difference_df["adjusted p-value"],
            rtol=0.01,
        )

    def test_multiple_difference_level_as_reference_false(self):
        difference_df = self.test.multiple_difference(level=("us", "control"), level_as_reference=False, verbose=True)
        assert len(difference_df) == (
            (self.data.variation_name.unique().size - 1) * self.data.country.unique().size
            + self.data.country.unique().size
            - 1
        )
        n_comp = len(difference_df)
        assert np.allclose(
            difference_df.apply(lambda row: min(row[P_VALUE] * n_comp, 1), axis=1),
            difference_df["adjusted p-value"],
            rtol=0.01,
        )

    def test_multiple_difference_groupby(self):
        difference_df = self.test.multiple_difference(
            level="control", groupby="country", level_as_reference=True, verbose=True
        )
        assert len(difference_df) == ((self.data.variation_name.unique().size - 1) * self.data.country.unique().size)
        n_comp = len(difference_df)
        assert np.allclose(
            difference_df.apply(lambda row: min(row[P_VALUE] * n_comp, 1), axis=1),
            difference_df["adjusted p-value"],
            rtol=0.01,
        )

    def test_summary_plot(self):
        chart_grid = self.test.summary_plot()
        assert len(chart_grid.charts) == 1

    def test_summary_plot_groupby(self):
        chart_grid = self.test.summary_plot(groupby="country")
        assert len(chart_grid.charts) == self.data.country.unique().size

    def test_difference_plot(self):
        chartgrid = self.test.difference_plot(level_1=("us", "control"), level_2=("ca", "test"))
        assert len(chartgrid.charts) == 1

    def test_difference_plot_groupby(self):
        chartgrid = self.test.difference_plot(level_1="control", level_2="test", groupby="country")
        assert len(chartgrid.charts) == 1

    def test_multiple_difference_plot(self):
        chartgrid = self.test.multiple_difference_plot(level=("us", "control"), level_as_reference=True)
        assert len(chartgrid.charts) == 1

    def test_multiple_difference_plot_groupby(self):
        chartgrid = self.test.multiple_difference_plot(level="control", groupby="country", level_as_reference=True)
        assert len(chartgrid.charts) == 1


class TestCategoricalContinuous(object):
    def setup(self):
        np.random.seed(123)

        self.data = pd.DataFrame(
            {
                "variation_name": ["test", "control", "test2", "test", "control", "test2"],
                "nr_of_items": [1969, 312, 2955, 195, 24, 330],
                "nr_of_items_sumsq": [5767, 984, 8771, 553, 80, 1010],
                "users": [1009, 104, 1502, 100, 10, 150],
                "country": ["us", "us", "us", "gb", "gb", "gb"],
            }
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns=["country", "variation_name"],
        )

    def test_init_one_country(self):
        spotify_confidence.ZTest(
            self.data.query('country == "us"'),
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns="variation_name",
        )

    def test_summary(self):
        summary_df = self.test.summary()
        assert len(summary_df) == len(self.data)

    def test_difference(self):
        difference_df = self.test.difference(level_1=("us", "control"), level_2=("us", "test"))
        assert len(difference_df) == 1

    def test_difference_groupby(self):
        difference_df = self.test.difference(level_1="control", level_2="test", groupby="country")
        assert len(difference_df) == self.data.country.unique().size

    def test_multiple_difference(self):
        difference_df = self.test.multiple_difference(level=("us", "control"), level_as_reference=True)
        assert len(difference_df) == (
            (self.data.variation_name.unique().size - 1) * self.data.country.unique().size
            + self.data.country.unique().size
            - 1
        )
        n_comp = len(difference_df)
        assert np.allclose(
            difference_df["p-value"].map(lambda p: min(1, n_comp * p)), difference_df["adjusted p-value"], rtol=0.01
        )

    def test_multiple_difference_groupby(self):
        difference_df = self.test.multiple_difference(level="control", groupby="country", level_as_reference=True)
        assert len(difference_df) == ((self.data.variation_name.unique().size - 1) * self.data.country.unique().size)
        n_comp = len(difference_df)
        assert np.allclose(
            difference_df["p-value"].map(lambda p: min(1, n_comp * p)), difference_df["adjusted p-value"], rtol=0.01
        )

    def test_summary_plot(self):
        chart_grid = self.test.summary_plot()
        assert len(chart_grid.charts) == 1

    def test_summary_plot_groupby(self):
        chart_grid = self.test.summary_plot(groupby="country")
        assert len(chart_grid.charts) == self.data.country.unique().size

    def test_difference_plot(self):
        chartgrid = self.test.difference_plot(level_1=("us", "control"), level_2=("gb", "test"))
        assert len(chartgrid.charts) == 1

    def test_difference_plot_groupby(self):
        chartgrid = self.test.difference_plot(level_1="control", level_2="test", groupby="country")
        assert len(chartgrid.charts) == 1

    def test_multiple_difference_plot(self):
        chartgrid = self.test.multiple_difference_plot(level=("us", "control"), level_as_reference=True)
        assert len(chartgrid.charts) == 1

    def test_multiple_difference_plot_groupby(self):
        chartgrid = self.test.multiple_difference_plot(level="control", groupby="country", level_as_reference=True)
        assert len(chartgrid.charts) == 1


class TestOrdinal(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                ],
                "nr_of_items": [500, 8, 100, 510, 8, 100, 520, 9, 104, 530, 7, 100, 530, 8, 103],
                "nr_of_items_sumsq": [2500, 12, 150, 2510, 13, 140, 2520, 14, 154, 2530, 15, 160, 2530, 16, 103],
                "users": [1010, 22, 150, 1000, 20, 153, 1030, 23, 154, 1000, 20, 150, 1040, 21, 155],
                "days_since_reg": [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5],
            }
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns="variation_name",
            ordinal_group_column="days_since_reg",
        )

    def test_summary(self):
        summary_df = self.test.summary()
        assert len(summary_df) == len(self.data)

    def test_difference(self):
        difference_df = self.test.difference(level_1=("control", 1), level_2=("test", 1))
        assert len(difference_df) == 1

    def test_difference_groupby(self):
        difference_df = self.test.difference(level_1="control", level_2="test", groupby="days_since_reg")
        assert len(difference_df) == self.data.days_since_reg.unique().size

    def test_multiple_difference(self):
        difference_df = self.test.multiple_difference(level=("control", 1), level_as_reference=True, verbose=True)
        assert len(difference_df) == (
            (self.data.variation_name.unique().size - 1) * self.data.days_since_reg.unique().size
            + self.data.days_since_reg.unique().size
            - 1
        )
        n_comp = len(difference_df)
        assert np.allclose(
            difference_df.apply(lambda row: min(row[P_VALUE] * n_comp, 1), axis=1),
            difference_df["adjusted p-value"],
            rtol=0.01,
        )

    def test_multiple_difference_groupby(self):
        difference_df = self.test.multiple_difference(
            level="control", groupby="days_since_reg", level_as_reference=True, verbose=True
        )
        assert len(difference_df) == (
            (self.data.variation_name.unique().size - 1) * self.data.days_since_reg.unique().size
        )
        n_comp = len(difference_df)
        assert np.allclose(
            difference_df.apply(lambda row: min(row[P_VALUE] * n_comp, 1), axis=1),
            difference_df["adjusted p-value"],
            rtol=0.01,
        )

    def test_summary_plot(self):
        chart_grid = self.test.summary_plot()
        assert len(chart_grid.charts) == 1

    def test_summary_plot_groupby(self):
        chart_grid = self.test.summary_plot(groupby="days_since_reg")
        assert len(chart_grid.charts) == self.data.days_since_reg.unique().size

    def test_difference_plot(self):
        chartgrid = self.test.difference_plot(level_1=("control", 1), level_2=("test", 2))
        assert len(chartgrid.charts) == 1

    def test_difference_plot_groupby(self):
        chartgrid = self.test.difference_plot(level_1="control", level_2="test", groupby="days_since_reg")
        assert len(chartgrid.charts) == 1

    def test_multiple_difference_plot(self):
        chartgrid = self.test.multiple_difference_plot(level=("control", 1), level_as_reference=True)
        assert len(chartgrid.charts) == 1

    def test_multiple_difference_plot_groupby(self):
        chartgrid = self.test.multiple_difference_plot(
            level="control", groupby="days_since_reg", level_as_reference=True
        )
        assert len(chartgrid.charts) == 1


class TestOrdinalPlusTwoCategorical(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                ],
                "nr_of_items": [
                    500,
                    8,
                    100,
                    510,
                    8,
                    100,
                    520,
                    9,
                    104,
                    530,
                    7,
                    100,
                    530,
                    8,
                    103,
                    500,
                    8,
                    100,
                    510,
                    8,
                    100,
                    520,
                    9,
                    104,
                    530,
                    7,
                    100,
                    530,
                    8,
                    103,
                ],
                "nr_of_items_sumsq": [
                    1010,
                    32,
                    250,
                    1000,
                    30,
                    253,
                    1030,
                    33,
                    254,
                    1000,
                    30,
                    250,
                    1040,
                    31,
                    255,
                    1010,
                    22,
                    150,
                    1000,
                    20,
                    153,
                    1030,
                    23,
                    154,
                    1000,
                    20,
                    150,
                    1040,
                    21,
                    155,
                ],
                "users": [
                    2010,
                    42,
                    250,
                    2000,
                    40,
                    253,
                    2030,
                    43,
                    254,
                    2000,
                    40,
                    250,
                    2040,
                    41,
                    255,
                    1010,
                    22,
                    150,
                    1000,
                    20,
                    153,
                    1030,
                    23,
                    154,
                    1000,
                    20,
                    150,
                    1040,
                    21,
                    155,
                ],
                "days_since_reg": [
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    3,
                    3,
                    3,
                    4,
                    4,
                    4,
                    5,
                    5,
                    5,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    3,
                    3,
                    3,
                    4,
                    4,
                    4,
                    5,
                    5,
                    5,
                ],
                "country": [
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                ],
                "non_inferiority_margin": [
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.01,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                ],
                "preferred_direction": [
                    DECREASE_PREFFERED,
                    DECREASE_PREFFERED,
                    DECREASE_PREFFERED,
                    DECREASE_PREFFERED,
                    DECREASE_PREFFERED,
                    DECREASE_PREFFERED,
                    DECREASE_PREFFERED,
                    DECREASE_PREFFERED,
                    DECREASE_PREFFERED,
                    DECREASE_PREFFERED,
                    DECREASE_PREFFERED,
                    DECREASE_PREFFERED,
                    DECREASE_PREFFERED,
                    DECREASE_PREFFERED,
                    DECREASE_PREFFERED,
                    INCREASE_PREFFERED,
                    INCREASE_PREFFERED,
                    INCREASE_PREFFERED,
                    INCREASE_PREFFERED,
                    INCREASE_PREFFERED,
                    INCREASE_PREFFERED,
                    INCREASE_PREFFERED,
                    INCREASE_PREFFERED,
                    INCREASE_PREFFERED,
                    INCREASE_PREFFERED,
                    INCREASE_PREFFERED,
                    INCREASE_PREFFERED,
                    INCREASE_PREFFERED,
                    INCREASE_PREFFERED,
                    INCREASE_PREFFERED,
                ],
            }
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns=["variation_name", "country"],
            ordinal_group_column="days_since_reg",
        )

    def test_summary(self):
        summary_df = self.test.summary()
        assert len(summary_df) == len(self.data)

    @pytest.mark.parametrize("correction_method", CORRECTION_METHODS, ids=lambda x: f"correction method: {x}")
    def test_difference(self, correction_method):
        self.test._confidence_computer._correction_method = correction_method
        if BONFERRONI in correction_method:
            difference_df = self.test.difference(level_1=("control", "gb", 1), level_2=("test", "us", 2))
            assert len(difference_df) == 1
        else:
            difference_df = self.test.difference(
                level_1=("control", "gb", 1), level_2=("test", "us", 2), non_inferiority_margins=(None, "increase")
            )
            assert len(difference_df) == 1

    @pytest.mark.parametrize("correction_method", CORRECTION_METHODS, ids=lambda x: f"correction method: {x}")
    def test_difference_groupby(self, correction_method):
        self.test._confidence_computer._correction_method = correction_method
        if BONFERRONI in correction_method:
            difference_df = self.test.difference(
                level_1="control", level_2="test", groupby=["country", "days_since_reg"]
            )
            assert len(difference_df) == self.data.days_since_reg.unique().size * self.data.country.unique().size
        else:
            difference_df = self.test.difference(
                level_1="control",
                level_2="test",
                groupby=["country", "days_since_reg"],
                non_inferiority_margins=(None, "increase"),
            )
            assert len(difference_df) == self.data.days_since_reg.unique().size * self.data.country.unique().size

    @pytest.mark.parametrize("correction_method", CORRECTION_METHODS, ids=lambda x: f"correction method: {x}")
    def test_multiple_difference(self, correction_method):
        self.test._confidence_computer._correction_method = correction_method
        if BONFERRONI in correction_method:
            difference_df = self.test.multiple_difference(
                level=("control", 1), groupby="country", level_as_reference=True
            )
            assert len(difference_df) == (
                self.data.country.unique().size
                * (
                    (self.data.variation_name.unique().size - 1) * self.data.days_since_reg.unique().size
                    + self.data.days_since_reg.unique().size
                    - 1
                )
            )
            n_comp = len(difference_df)
            assert np.allclose(
                difference_df["p-value"].map(lambda p: min(1, n_comp * p)),
                difference_df["adjusted p-value"],
                rtol=0.01,
            )
        else:
            difference_df = self.test.multiple_difference(
                level=("control", 1),
                groupby="country",
                level_as_reference=True,
                non_inferiority_margins=(None, "increase"),
            )
            assert len(difference_df) == (
                self.data.country.unique().size
                * (
                    (self.data.variation_name.unique().size - 1) * self.data.days_since_reg.unique().size
                    + self.data.days_since_reg.unique().size
                    - 1
                )
            )

    @pytest.mark.parametrize("correction_method", CORRECTION_METHODS, ids=lambda x: f"correction method: {x}")
    def test_multiple_difference_groupby(self, correction_method):
        self.test._confidence_computer._correction_method = correction_method
        if BONFERRONI in correction_method:
            difference_df = self.test.multiple_difference(
                level="control", groupby=["days_since_reg", "country"], level_as_reference=True
            )
            assert len(difference_df) == (
                (self.data.variation_name.unique().size - 1)
                * self.data.days_since_reg.unique().size
                * self.data.country.unique().size
            )

            n_comp = len(difference_df)
            assert np.allclose(
                difference_df["p-value"].map(lambda p: min(1, n_comp * p)),
                difference_df["adjusted p-value"],
                rtol=0.01,
            )
        else:
            difference_df = self.test.multiple_difference(
                level="control",
                groupby=["days_since_reg", "country"],
                level_as_reference=True,
                non_inferiority_margins=(None, "increase"),
            )
            assert len(difference_df) == (
                (self.data.variation_name.unique().size - 1)
                * self.data.days_since_reg.unique().size
                * self.data.country.unique().size
            )
            if correction_method in CORRECTION_METHODS_THAT_SUPPORT_CI:
                assert not any(difference_df[ADJUSTED_LOWER].isna())

    @pytest.mark.parametrize("correction_method", CORRECTION_METHODS, ids=lambda x: f"correction method: {x}")
    def test_differece_with_nims(self, correction_method):
        self.test._confidence_computer._correction_method = correction_method
        df = self.test.difference(
            level_1=("test", "us"),
            level_2=("control", "us"),
            groupby="days_since_reg",
            non_inferiority_margins=(0.01, INCREASE_PREFFERED),
        )
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(
            level_1=("test", "us"),
            level_2=("control", "us"),
            groupby=["days_since_reg"],
            non_inferiority_margins=(0.01, DECREASE_PREFFERED),
        )
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(
            level_1=("test", 1),
            level_2=("control", 1),
            groupby=["country"],
            non_inferiority_margins={"us": (0.01, INCREASE_PREFFERED), "gb": (0.05, INCREASE_PREFFERED)},
        )
        assert len(df) == 2
        assert "country" in df.columns

        df = self.test.difference(
            level_1="test",
            level_2="control",
            groupby=["country", "days_since_reg"],
            non_inferiority_margins=(0.01, DECREASE_PREFFERED),
        )
        assert len(df) == 10
        assert "country" in df.columns
        assert "days_since_reg" in df.columns

        nims = {
            ("us", 1): (0.01, DECREASE_PREFFERED),
            ("us", 2): (0.1, INCREASE_PREFFERED),
            ("us", 3): (0.2, DECREASE_PREFFERED),
            ("us", 4): (0.5, INCREASE_PREFFERED),
            ("us", 5): (0.99, DECREASE_PREFFERED),
            ("gb", 1): (1.01, INCREASE_PREFFERED),
            ("gb", 2): (2.01, DECREASE_PREFFERED),
            ("gb", 3): (3.01, INCREASE_PREFFERED),
            ("gb", 4): (4.01, DECREASE_PREFFERED),
            ("gb", 5): (5.01, INCREASE_PREFFERED),
        }
        df = self.test.difference(
            level_1="test", level_2="control", groupby=["country", "days_since_reg"], non_inferiority_margins=nims
        )
        assert len(df) == 10
        assert "country" in df.columns
        assert "days_since_reg" in df.columns

        df = self.test.multiple_difference(
            level="control",
            level_as_reference=True,
            groupby=["country", "days_since_reg"],
            non_inferiority_margins=nims,
        )
        assert len(df) == 20

    @pytest.mark.parametrize("correction_method", CORRECTION_METHODS, ids=lambda x: f"correction method: {x}")
    def test_differece_with_nims_in_df(self, correction_method):
        self.test._confidence_computer._correction_method = correction_method

        df = self.test.difference(
            level_1=("test", "us"), level_2=("control", "us"), groupby="days_since_reg", non_inferiority_margins=True
        )
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(
            level_1=("test", "us"), level_2=("control", "us"), groupby=["days_since_reg"], non_inferiority_margins=True
        )
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(
            level_1=("test", 1), level_2=("control", 1), groupby=["country"], non_inferiority_margins=True
        )
        assert len(df) == 2
        assert "country" in df.columns

        df = self.test.difference(
            level_1="test", level_2="control", groupby=["country", "days_since_reg"], non_inferiority_margins=True
        )
        assert len(df) == 10
        assert "country" in df.columns
        assert "days_since_reg" in df.columns

        df = self.test.difference(
            level_1="test", level_2="control", groupby=["country", "days_since_reg"], non_inferiority_margins=True
        )
        assert len(df) == 10
        assert "country" in df.columns
        assert "days_since_reg" in df.columns

        df = self.test.multiple_difference(
            level="control",
            level_as_reference=True,
            groupby=["country", "days_since_reg"],
            non_inferiority_margins=True,
        )
        assert len(df) == 20

    def test_summary_plot(self):
        chart_grid = self.test.summary_plot()
        assert len(chart_grid.charts) == 1

    def test_summary_plot_groupby(self):
        chart_grid = self.test.summary_plot(groupby="country")
        assert len(chart_grid.charts) == self.data.country.unique().size

    def test_summary_plot_groupby_2(self):
        chart_grid = self.test.summary_plot(groupby=["days_since_reg", "country"])
        assert len(chart_grid.charts) == (self.data.country.unique().size * self.data.days_since_reg.unique().size)

    def test_difference_plot(self):
        chartgrid = self.test.difference_plot(level_1=("control", "gb", 1), level_2=("test", "us", 2))
        assert len(chartgrid.charts) == 1

    def test_difference_plot_groupby(self):
        chartgrid = self.test.difference_plot(
            level_1=("control", "gb"), level_2=("test", "us"), groupby="days_since_reg"
        )
        assert len(chartgrid.charts) == 1

    def test_difference_plot_groupby_2(self):
        chartgrid = self.test.difference_plot(level_1="control", level_2="test", groupby=["days_since_reg", "country"])
        assert len(chartgrid.charts) == 1

    def test_multiple_difference_plot(self):
        chartgrid = self.test.multiple_difference_plot(
            level=("control", 1), groupby="country", level_as_reference=True
        )
        assert len(chartgrid.charts) == 1

    def test_multiple_difference_plot_groupby(self):
        chartgrid = self.test.multiple_difference_plot(
            level="control", groupby=["days_since_reg", "country"], level_as_reference=True
        )
        assert len(chartgrid.charts) == 1

    def test_differece_plot_with_nims(self):
        ch = self.test.difference_plot(
            level_1=("test", "us"),
            level_2=("control", "us"),
            groupby="days_since_reg",
            non_inferiority_margins=(0.01, INCREASE_PREFFERED),
        )
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(
            level_1=("test", "us"),
            level_2=("control", "us"),
            groupby=["days_since_reg"],
            non_inferiority_margins=(0.01, DECREASE_PREFFERED),
        )
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(
            level_1=("test", 1),
            level_2=("control", 1),
            groupby=["country"],
            non_inferiority_margins={"us": (0.01, INCREASE_PREFFERED), "gb": (0.05, INCREASE_PREFFERED)},
        )
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(
            level_1="test",
            level_2="control",
            groupby=["country", "days_since_reg"],
            non_inferiority_margins=(0.01, DECREASE_PREFFERED),
        )
        assert len(ch.charts) == 1

        nims = {
            ("us", 1): (0.01, DECREASE_PREFFERED),
            ("us", 2): (0.1, INCREASE_PREFFERED),
            ("us", 3): (0.2, DECREASE_PREFFERED),
            ("us", 4): (0.5, INCREASE_PREFFERED),
            ("us", 5): (0.99, DECREASE_PREFFERED),
            ("gb", 1): (1.01, INCREASE_PREFFERED),
            ("gb", 2): (2.01, DECREASE_PREFFERED),
            ("gb", 3): (3.01, INCREASE_PREFFERED),
            ("gb", 4): (4.01, DECREASE_PREFFERED),
            ("gb", 5): (5.01, INCREASE_PREFFERED),
        }
        ch = self.test.difference_plot(
            level_1="test", level_2="control", groupby=["country", "days_since_reg"], non_inferiority_margins=nims
        )
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot(
            level="control",
            level_as_reference=True,
            groupby=["country", "days_since_reg"],
            non_inferiority_margins=nims,
        )
        assert len(ch.charts) == 1

    def test_differece_plot_with_nims_in_df(self):
        ch = self.test.difference_plot(
            level_1=("test", "us"), level_2=("control", "us"), groupby="days_since_reg", non_inferiority_margins=True
        )
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(
            level_1=("test", "us"), level_2=("control", "us"), groupby=["days_since_reg"], non_inferiority_margins=True
        )
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(
            level_1=("test", 1), level_2=("control", 1), groupby=["country"], non_inferiority_margins=True
        )
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(
            level_1="test", level_2="control", groupby=["country", "days_since_reg"], non_inferiority_margins=True
        )
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(
            level_1="test", level_2="control", groupby=["country", "days_since_reg"], non_inferiority_margins=True
        )
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot(
            level="control",
            level_as_reference=True,
            groupby=["country", "days_since_reg"],
            non_inferiority_margins=True,
        )
        assert len(ch.charts) == 1


class TestCategoricalBinomialData(object):
    def setup(self):
        np.random.seed(123)

        self.data = pd.DataFrame(
            {
                "variation_name": ["test", "control", "test2", "test", "control", "test2"],
                "success": [500, 42, 1005, 50, 4, 100],
                "total": [1009, 104, 1502, 100, 10, 150],
                "country": [
                    "us",
                    "us",
                    "us",
                    "gb",
                    "gb",
                    "gb",
                ],
            }
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="success",
            numerator_sum_squares_column="success",
            denominator_column="total",
            categorical_group_columns=["country", "variation_name"],
        )

    def test_summary(self):
        """Area plot tests"""

        summary = self.test.summary()
        assert np.array_equal(summary.country, np.array(["us", "us", "us", "gb", "gb", "gb"]))
        assert np.array_equal(summary.point_estimate, self.data.success / self.data.total)
        assert np.allclose(
            summary["ci_lower"],
            np.array(
                [
                    0.4646901340180582,
                    0.30954466010970333,
                    0.6453118311511006,
                    0.4020018007729973,
                    0.0963636851484016,
                    0.5912276177282552,
                ]
            ),
            rtol=1e-3,
        )
        assert np.allclose(
            summary["ci_upper"],
            np.array(
                [
                    0.5263901434844195,
                    0.4981476475826044,
                    0.692903881232388,
                    0.5979981992270027,
                    0.7036363148515985,
                    0.7421057156050781,
                ]
            ),
            rtol=1e-3,
        )

    def test_multiple_difference(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference(("bad_value", "bad_value"), level_as_reference=False)

        diff = self.test.multiple_difference(("us", "control"), level_as_reference=False)
        assert np.allclose(
            diff["adjusted p-value"],
            np.array([1e00, 8.291843e-01, 9.971992e-05, 3.504662e-01, 4.504966e-07]),
            rtol=1e-3,
        )
        assert np.allclose(
            diff["p-value"],
            np.array([9.81084197e-01, 1.65836862e-01, 1.99439850e-05, 7.00932382e-02, 9.00993166e-08]),
            rtol=1e-3,
        )
        assert np.allclose(
            diff["adjusted ci_lower"],
            np.array([-0.41400184, -0.27489017, -0.42153065, -0.22209041, -0.39307973]),
            rtol=1e-3,
        )
        assert np.allclose(
            diff["adjusted ci_upper"],
            np.array([0.42169415, 0.08258247, -0.10411038, 0.03870244, -0.13744367]),
            rtol=1e-3,
        )

        diff = self.test.multiple_difference("test", groupby="country", level_as_reference=False)
        assert np.allclose(
            diff["adjusted p-value"],
            np.array([1.00000000e00, 3.30302805e-02, 2.80372953e-01, 0.0]),
            rtol=1e-3,
        )
        assert np.allclose(
            diff["p-value"],
            np.array([5.39020329e-01, 8.25757011e-03, 7.00932382e-02, 0.0]),
            rtol=1e-3,
        )
        assert np.allclose(
            diff["adjusted ci_lower"],
            np.array([-0.30659699, -0.32426934, -0.03474758, -0.2232184]),
            rtol=1e-3,
        )
        assert np.allclose(
            diff["adjusted ci_upper"],
            np.array([0.50659699, -0.00906399, 0.21813554, -0.12391703]),
            rtol=1e-3,
        )


class TestWithNims(object):
    def setup(self):
        self.data = pd.DataFrame(
            [
                {
                    "group": "1",
                    "count": 5000,
                    "sum": 10021.0,
                    "sum_of_squares": 25142.0,
                    "avg": 2.004210,
                    "var": 1.0116668,
                },
                {
                    "group": "2",
                    "count": 5000,
                    "sum": 9892.0,
                    "sum_of_squares": 24510.0,
                    "avg": 1.978424,
                    "var": 0.9881132,
                },
            ]
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="sum",
            numerator_sum_squares_column="sum_of_squares",
            denominator_column="count",
            categorical_group_columns="group",
            interval_size=0.99,
        )

    @pytest.mark.parametrize("correction_method", CORRECTION_METHODS, ids=lambda x: f"correction method: {x}")
    def test_compare_series_non_inferiority_improve_postitive(self, correction_method):
        summary = self.test.summary()
        control_avg = self.data.query("group == '1'").avg.values[0]
        assert np.allclose(control_avg, summary.query("group == '1'")[POINT_ESTIMATE])

        diff = self.test.difference(level_1="1", level_2="2", non_inferiority_margins=(0.02, "increase"))

        np.testing.assert_almost_equal(diff[DIFFERENCE].values[0], -0.0258, 3)
        assert np.isinf(diff[CI_UPPER].values[0])
        np.testing.assert_almost_equal(diff[CI_LOWER].values[0], -0.0723, 3)
        assert diff[P_VALUE].values[0] > 0.01

        self.test._confidence_computer._correction_method = correction_method
        diff_2 = self.test.difference(level_1="1", level_2="2", non_inferiority_margins=(0.02, "increase"))

        if SPOT_1 in correction_method:
            assert all(diff[CI_LOWER] == diff_2[CI_LOWER])
            assert np.isfinite(diff_2[CI_UPPER].values[0])

        if BONFERRONI not in correction_method and correction_method in CORRECTION_METHODS_THAT_SUPPORT_CI:
            assert all(diff[ADJUSTED_LOWER] <= diff_2[ADJUSTED_LOWER])
            assert all(diff[ADJUSTED_UPPER] >= diff_2[ADJUSTED_UPPER])

    @pytest.mark.parametrize("correction_method", CORRECTION_METHODS, ids=lambda x: f"correction method: {x}")
    def test_compare_series_non_inferiority_improve_negative(self, correction_method):
        summary = self.test.summary()
        control_avg = self.data.query("group == '1'").avg.values[0]
        assert np.allclose(control_avg, summary.query("group == '1'")[POINT_ESTIMATE])

        diff = self.test.difference(level_1="1", level_2="2", non_inferiority_margins=(0.02, "decrease"))

        np.testing.assert_almost_equal(diff[DIFFERENCE].values[0], -0.0258, 3)
        assert diff[CI_LOWER].values[0] == -float("inf")
        np.testing.assert_almost_equal(diff[CI_UPPER].values[0], 0.0207, 3)
        assert diff[P_VALUE].values[0] < 0.01

        self.test._confidence_computer._correction_method = correction_method
        diff_2 = self.test.difference(level_1="1", level_2="2", non_inferiority_margins=(0.02, "decrease"))

        if SPOT_1 in correction_method:
            assert all(diff[CI_UPPER] == diff_2[CI_UPPER])
            assert np.isfinite(diff_2[CI_LOWER].values[0])

        if BONFERRONI not in correction_method and correction_method in CORRECTION_METHODS_THAT_SUPPORT_CI:
            assert all(diff[ADJUSTED_LOWER] <= diff_2[ADJUSTED_LOWER])
            assert all(diff[ADJUSTED_UPPER] >= diff_2[ADJUSTED_UPPER])

    def test_one_sided_ztest_positive(self):
        summary = self.test.summary()
        control_avg = self.data.query("group == '1'").avg.values[0]
        assert np.allclose(control_avg, summary.query("group == '1'")[POINT_ESTIMATE])

        diff = self.test.difference(level_1="1", level_2="2", non_inferiority_margins=(None, "increase"))

        np.testing.assert_almost_equal(diff[DIFFERENCE].values[0], -0.0258, 3)
        assert diff[CI_UPPER].values[0] == float("inf")
        np.testing.assert_almost_equal(diff[CI_LOWER].values[0], -0.0723, 3)
        assert diff[P_VALUE].values[0] > 0.01

    def test_one_sided_ztest_negative(self):
        summary = self.test.summary()
        control_avg = self.data.query("group == '1'").avg.values[0]
        assert np.allclose(control_avg, summary.query("group == '1'")[POINT_ESTIMATE])

        diff = self.test.difference(level_1="1", level_2="2", non_inferiority_margins=(None, "decrease"))

        np.testing.assert_almost_equal(diff[DIFFERENCE].values[0], -0.0258, 3)
        assert diff[CI_LOWER].values[0] == -float("inf")
        np.testing.assert_almost_equal(diff[CI_UPPER].values[0], 0.0207, 3)
        assert diff[P_VALUE].values[0] > 0.01


class TestSequentialOrdinalPlusTwoCategorical(object):
    def setup(self):
        np.random.seed(123)
        d = 50 + 1 * np.random.randn(60)
        u = np.floor(2000 + np.linspace(0, 1000, 60) + 10 * np.random.randn(60))
        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                ],
                "nr_of_items": d,
                "nr_of_items_sumsq": d / 20,
                "users": u,
                "date": pd.to_datetime(
                    [
                        "2021-04-01",
                        "2021-04-01",
                        "2021-04-01",
                        "2021-04-02",
                        "2021-04-02",
                        "2021-04-02",
                        "2021-04-03",
                        "2021-04-03",
                        "2021-04-03",
                        "2021-04-04",
                        "2021-04-04",
                        "2021-04-04",
                        "2021-04-05",
                        "2021-04-05",
                        "2021-04-05",
                        "2021-04-01",
                        "2021-04-01",
                        "2021-04-01",
                        "2021-04-02",
                        "2021-04-02",
                        "2021-04-02",
                        "2021-04-03",
                        "2021-04-03",
                        "2021-04-03",
                        "2021-04-04",
                        "2021-04-04",
                        "2021-04-04",
                        "2021-04-05",
                        "2021-04-05",
                        "2021-04-05",
                        "2021-04-01",
                        "2021-04-01",
                        "2021-04-01",
                        "2021-04-02",
                        "2021-04-02",
                        "2021-04-02",
                        "2021-04-03",
                        "2021-04-03",
                        "2021-04-03",
                        "2021-04-04",
                        "2021-04-04",
                        "2021-04-04",
                        "2021-04-05",
                        "2021-04-05",
                        "2021-04-05",
                        "2021-04-01",
                        "2021-04-01",
                        "2021-04-01",
                        "2021-04-02",
                        "2021-04-02",
                        "2021-04-02",
                        "2021-04-03",
                        "2021-04-03",
                        "2021-04-03",
                        "2021-04-04",
                        "2021-04-04",
                        "2021-04-04",
                        "2021-04-05",
                        "2021-04-05",
                        "2021-04-05",
                    ]
                ),
                "country": [
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                ],
                "metric": [
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m1",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                    "m2",
                ],
                "non_inferiority_margin": [
                    0.2,
                    0.2,
                    0.2,
                    0.2,
                    0.2,
                    0.2,
                    0.2,
                    0.2,
                    0.2,
                    0.2,
                    0.2,
                    0.2,
                    0.2,
                    0.2,
                    0.2,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0.1,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                ],
                "preferred_direction": [
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    "increase",
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                ],
            }
        ).assign(final_sample_size=5000)

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns=["variation_name", "country", "metric"],
            ordinal_group_column="date",
        )

    @pytest.mark.parametrize("correction_method", CORRECTION_METHODS, ids=lambda x: f"correction method: {x}")
    def test_multiple_difference_groupby(self, correction_method):
        self.test._confidence_computer._correction_method = correction_method

        def get_diff() -> pd.DataFrame:
            return self.test.multiple_difference(
                level="control",
                groupby=["date", "country", "metric"],
                level_as_reference=True,
                final_expected_sample_size_column="final_sample_size",
            )

        if BONFERRONI not in correction_method:
            with pytest.raises(ValueError):
                difference_df = get_diff()
        else:
            difference_df = get_diff()

            assert len(difference_df) == (
                (self.data.variation_name.unique().size - 1)
                * self.data.date.unique().size
                * self.data.country.unique().size
                * self.data.metric.unique().size
            )
            assert difference_df["p-value"].isnull().all()
            assert difference_df["adjusted p-value"].isnull().all()

    def test_multiple_difference_plot_groupby(self):
        charts = self.test.multiple_difference_plot(
            level="control",
            groupby=["date", "country", "metric"],
            level_as_reference=True,
            final_expected_sample_size_column="final_sample_size",
        ).charts
        assert len(charts) == 1

    def test_multiple_difference_groupby_onesided_decrease(self):
        difference_df = self.test.multiple_difference(
            level="control",
            groupby=["date", "country", "metric"],
            level_as_reference=True,
            non_inferiority_margins=(0.05, "decrease"),
            final_expected_sample_size_column="final_sample_size",
        )
        assert len(difference_df) == (
            (self.data.variation_name.unique().size - 1)
            * self.data.date.unique().size
            * self.data.country.unique().size
            * self.data.metric.unique().size
        )
        assert difference_df["p-value"].isnull().all()
        assert difference_df["adjusted p-value"].isnull().all()

    def test_multiple_difference_groupby_onesided_increase(self):
        difference_df = self.test.multiple_difference(
            level="control",
            groupby=["date", "country", "metric"],
            level_as_reference=True,
            non_inferiority_margins=(0.05, "increase"),
            final_expected_sample_size_column="final_sample_size",
        )
        assert len(difference_df) == (
            (self.data.variation_name.unique().size - 1)
            * self.data.date.unique().size
            * self.data.country.unique().size
            * self.data.metric.unique().size
        )
        assert difference_df["p-value"].isnull().all()
        assert difference_df["adjusted p-value"].isnull().all()

    def test_multiple_difference_groupby_mixed_nims(self):
        nims = {
            (pd.to_datetime("2021-04-01"), "us", "m1"): (0.2, "increase"),
            (pd.to_datetime("2021-04-02"), "us", "m1"): (0.2, "increase"),
            (pd.to_datetime("2021-04-03"), "us", "m1"): (0.2, "increase"),
            (pd.to_datetime("2021-04-04"), "us", "m1"): (0.2, "increase"),
            (pd.to_datetime("2021-04-05"), "us", "m1"): (0.2, "increase"),
            (pd.to_datetime("2021-04-01"), "gb", "m1"): (0.1, "increase"),
            (pd.to_datetime("2021-04-02"), "gb", "m1"): (0.1, "increase"),
            (pd.to_datetime("2021-04-03"), "gb", "m1"): (0.1, "increase"),
            (pd.to_datetime("2021-04-04"), "gb", "m1"): (0.1, "increase"),
            (pd.to_datetime("2021-04-05"), "gb", "m1"): (0.1, "increase"),
            (pd.to_datetime("2021-04-01"), "us", "m2"): (0, None),
            (pd.to_datetime("2021-04-02"), "us", "m2"): (0, None),
            (pd.to_datetime("2021-04-03"), "us", "m2"): (0, None),
            (pd.to_datetime("2021-04-04"), "us", "m2"): (0, None),
            (pd.to_datetime("2021-04-05"), "us", "m2"): (0, None),
            (pd.to_datetime("2021-04-01"), "gb", "m2"): (0, None),
            (pd.to_datetime("2021-04-02"), "gb", "m2"): (0, None),
            (pd.to_datetime("2021-04-03"), "gb", "m2"): (0, None),
            (pd.to_datetime("2021-04-04"), "gb", "m2"): (0, None),
            (pd.to_datetime("2021-04-05"), "gb", "m2"): (0, None),
        }

        difference_df = self.test.multiple_difference(
            level="control",
            groupby=["date", "country", "metric"],
            level_as_reference=True,
            non_inferiority_margins=nims,
            final_expected_sample_size_column="final_sample_size",
        )
        assert len(difference_df) == (
            (self.data.variation_name.unique().size - 1)
            * self.data.date.unique().size
            * self.data.country.unique().size
            * self.data.metric.unique().size
        )
        assert difference_df["p-value"].isnull().all()
        assert difference_df["adjusted p-value"].isnull().all()

        difference_df_2 = self.test.multiple_difference(
            level="control",
            groupby=["date", "country", "metric"],
            level_as_reference=True,
            non_inferiority_margins=True,
            final_expected_sample_size_column="final_sample_size",
        )

        for column in difference_df.columns:
            assert (difference_df[column] == difference_df_2[column]).all() or (
                difference_df["p-value"].isnull() == difference_df_2["p-value"].isnull()
            ).all()


DATE = "date"
COUNT = "count"
SUM = "sum"
SUM_OF_SQUARES = "sum_of_squares"
GROUP = "group"


class TestSequentialOrdinalPlusTwoCategorical2(object):
    def setup(self):
        self.data = pd.DataFrame(
            [
                {
                    DATE: "2020-04-01",
                    GROUP: "1",
                    "country": "swe",
                    "platform": "ios",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1000,
                    SUM: 2016.416,
                    SUM_OF_SQUARES: 5082.122,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-01",
                    GROUP: "2",
                    "country": "swe",
                    "platform": "ios",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1000,
                    SUM: 2028.478,
                    SUM_OF_SQUARES: 5210.193,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-01",
                    GROUP: "1",
                    "country": "fin",
                    "platform": "ios",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1000,
                    SUM: 1991.554,
                    SUM_OF_SQUARES: 4919.282,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-01",
                    GROUP: "2",
                    "country": "fin",
                    "platform": "ios",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1000,
                    SUM: 1958.713,
                    SUM_OF_SQUARES: 4818.665,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-01",
                    GROUP: "1",
                    "country": "swe",
                    "platform": "andr",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1000,
                    SUM: 2030.252,
                    SUM_OF_SQUARES: 5129.574,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-01",
                    GROUP: "2",
                    "country": "swe",
                    "platform": "andr",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1000,
                    SUM: 1966.138,
                    SUM_OF_SQUARES: 4848.321,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-01",
                    GROUP: "1",
                    "country": "fin",
                    "platform": "andr",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1000,
                    SUM: 1995.389,
                    SUM_OF_SQUARES: 4992.710,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-01",
                    GROUP: "2",
                    "country": "fin",
                    "platform": "andr",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1000,
                    SUM: 1952.098,
                    SUM_OF_SQUARES: 4798.772,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "1",
                    "country": "swe",
                    "platform": "ios",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1500,
                    SUM: 2986.667,
                    SUM_OF_SQUARES: 7427.582,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "2",
                    "country": "swe",
                    "platform": "ios",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1500,
                    SUM: 2989.488,
                    SUM_OF_SQUARES: 7421.710,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "1",
                    "country": "fin",
                    "platform": "ios",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1500,
                    SUM: 3008.681,
                    SUM_OF_SQUARES: 7565.406,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "2",
                    "country": "fin",
                    "platform": "ios",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1500,
                    SUM: 2933.173,
                    SUM_OF_SQUARES: 7207.038,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "1",
                    "country": "swe",
                    "platform": "andr",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1500,
                    SUM: 2986.308,
                    SUM_OF_SQUARES: 7584.148,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "2",
                    "country": "swe",
                    "platform": "andr",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1500,
                    SUM: 2985.802,
                    SUM_OF_SQUARES: 7446.539,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "1",
                    "country": "fin",
                    "platform": "andr",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1500,
                    SUM: 3008.190,
                    SUM_OF_SQUARES: 7532.521,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "2",
                    "country": "fin",
                    "platform": "andr",
                    "metric": "bananas_per_user_1d",
                    COUNT: 1500,
                    SUM: 3001.494,
                    SUM_OF_SQUARES: 7467.535,
                    "non_inferiority_margin": None,
                    "preferred_direction": None,
                },
                {
                    DATE: "2020-04-01",
                    GROUP: "1",
                    "country": "swe",
                    "platform": "ios",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1000,
                    SUM: 2016.416,
                    SUM_OF_SQUARES: 5082.122,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
                {
                    DATE: "2020-04-01",
                    GROUP: "2",
                    "country": "swe",
                    "platform": "ios",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1000,
                    SUM: 2028.478,
                    SUM_OF_SQUARES: 5210.193,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
                {
                    DATE: "2020-04-01",
                    GROUP: "1",
                    "country": "fin",
                    "platform": "ios",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1000,
                    SUM: 1991.554,
                    SUM_OF_SQUARES: 4919.282,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
                {
                    DATE: "2020-04-01",
                    GROUP: "2",
                    "country": "fin",
                    "platform": "ios",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1000,
                    SUM: 1958.713,
                    SUM_OF_SQUARES: 4818.665,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
                {
                    DATE: "2020-04-01",
                    GROUP: "1",
                    "country": "swe",
                    "platform": "andr",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1000,
                    SUM: 2030.252,
                    SUM_OF_SQUARES: 5129.574,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
                {
                    DATE: "2020-04-01",
                    GROUP: "2",
                    "country": "swe",
                    "platform": "andr",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1000,
                    SUM: 1966.138,
                    SUM_OF_SQUARES: 4848.321,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
                {
                    DATE: "2020-04-01",
                    GROUP: "1",
                    "country": "fin",
                    "platform": "andr",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1000,
                    SUM: 1995.389,
                    SUM_OF_SQUARES: 4992.710,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
                {
                    DATE: "2020-04-01",
                    GROUP: "2",
                    "country": "fin",
                    "platform": "andr",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1000,
                    SUM: 1952.098,
                    SUM_OF_SQUARES: 4798.772,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "1",
                    "country": "swe",
                    "platform": "ios",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1500,
                    SUM: 2986.667,
                    SUM_OF_SQUARES: 7427.582,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "2",
                    "country": "swe",
                    "platform": "ios",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1500,
                    SUM: 2989.488,
                    SUM_OF_SQUARES: 7421.710,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "1",
                    "country": "fin",
                    "platform": "ios",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1500,
                    SUM: 3008.681,
                    SUM_OF_SQUARES: 7565.406,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "2",
                    "country": "fin",
                    "platform": "ios",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1500,
                    SUM: 2933.173,
                    SUM_OF_SQUARES: 7207.038,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "1",
                    "country": "swe",
                    "platform": "andr",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1500,
                    SUM: 2986.308,
                    SUM_OF_SQUARES: 7584.148,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "2",
                    "country": "swe",
                    "platform": "andr",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1500,
                    SUM: 2985.802,
                    SUM_OF_SQUARES: 7446.539,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "1",
                    "country": "fin",
                    "platform": "andr",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1500,
                    SUM: 3008.190,
                    SUM_OF_SQUARES: 7532.521,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
                {
                    DATE: "2020-04-02",
                    GROUP: "2",
                    "country": "fin",
                    "platform": "andr",
                    "metric": "bananas_per_user_7d",
                    COUNT: 1500,
                    SUM: 3001.494,
                    SUM_OF_SQUARES: 7467.535,
                    "non_inferiority_margin": 0.01,
                    "preferred_direction": "increase",
                },
            ]
        )
        self.data[DATE] = pd.to_datetime(self.data[DATE])
        self.data = (
            self.data.groupby([DATE, GROUP, "country", "platform", "metric"])
            .sum(numeric_only=True)
            .groupby([GROUP, "country", "platform", "metric"])
            .cumsum()
            .reset_index()
            .assign(
                non_inferiority_margin=lambda df: df["metric"].map(
                    {"bananas_per_user_1d": None, "bananas_per_user_7d": 0.01}
                )
            )
            .assign(
                preferred_direction=lambda df: df["metric"].map(
                    {"bananas_per_user_1d": None, "bananas_per_user_7d": "increase"}
                )
            )
            .assign(final_expected_sample_size=5000)
        )

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column=SUM,
            numerator_sum_squares_column=SUM_OF_SQUARES,
            denominator_column=COUNT,
            categorical_group_columns=[GROUP, "country", "platform", "metric"],
            ordinal_group_column=DATE,
            interval_size=1 - 0.01,
            correction_method=BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY,
        )

    def test_with_manual_correction(self):
        test = spotify_confidence.ZTest(
            self.data.assign(blabla="hej"),
            numerator_column=SUM,
            numerator_sum_squares_column=SUM_OF_SQUARES,
            denominator_column=COUNT,
            categorical_group_columns=[GROUP, "country", "platform", "metric"],
            ordinal_group_column=DATE,
            interval_size=1 - 0.01 / 4,
        )

        difference_df = test.difference(
            level_1=("1", "fin", "andr", "bananas_per_user_1d"),
            level_2=("2", "fin", "andr", "bananas_per_user_1d"),
            groupby="date",
            final_expected_sample_size_column="final_expected_sample_size",
        )
        np.testing.assert_almost_equal(difference_df[ADJUSTED_LOWER].values[0], -0.2016570, 3)
        np.testing.assert_almost_equal(difference_df[ADJUSTED_UPPER].values[0], 0.11507406, 3)
        np.testing.assert_almost_equal(difference_df[ADJUSTED_LOWER].values[1], -0.1063633, 3)
        np.testing.assert_almost_equal(difference_df[ADJUSTED_UPPER].values[1], 0.06637345, 3)

        difference_df = test.difference(
            level_1=("1", "swe", "ios", "bananas_per_user_1d"),
            level_2=("2", "swe", "ios", "bananas_per_user_1d"),
            groupby="date",
            final_expected_sample_size_column="final_expected_sample_size",
        )
        np.testing.assert_almost_equal(difference_df[ADJUSTED_LOWER].values[0], -0.1506963, 3)
        np.testing.assert_almost_equal(difference_df[ADJUSTED_UPPER].values[0], 0.17481994, 3)
        np.testing.assert_almost_equal(difference_df[ADJUSTED_LOWER].values[1], -0.0812409, 3)
        np.testing.assert_almost_equal(difference_df[ADJUSTED_UPPER].values[1], 0.09314668, 3)

        difference_df = test.difference(
            level_1=("1", "fin", "andr", "bananas_per_user_7d"),
            level_2=("2", "fin", "andr", "bananas_per_user_7d"),
            groupby="date",
            non_inferiority_margins=(0.01, "increase"),
            final_expected_sample_size_column="final_expected_sample_size",
        )
        np.testing.assert_almost_equal(difference_df[ADJUSTED_LOWER].values[0], -0.1932786, 3)
        np.isinf(difference_df[ADJUSTED_UPPER].values[0])
        np.testing.assert_almost_equal(difference_df[ADJUSTED_LOWER].values[1], -0.10027731, 3)
        np.isinf(difference_df[ADJUSTED_UPPER].values[1])

        difference_df = test.difference(
            level_1=("1", "swe", "ios", "bananas_per_user_7d"),
            level_2=("2", "swe", "ios", "bananas_per_user_7d"),
            groupby="date",
            non_inferiority_margins=(0.01, "increase"),
            final_expected_sample_size_column="final_expected_sample_size",
        )
        np.testing.assert_almost_equal(difference_df[ADJUSTED_LOWER].values[0], -0.1420855, 3)
        np.isinf(difference_df[ADJUSTED_UPPER].values[0])
        np.testing.assert_almost_equal(difference_df[ADJUSTED_LOWER].values[1], -0.07509674, 3)
        np.isinf(difference_df[ADJUSTED_UPPER].values[1])

    def test_multiple_difference_plot(self):
        charts = self.test.multiple_difference_plot(
            level="1",
            groupby=["date", "country", "platform", "metric"],
            level_as_reference=True,
            final_expected_sample_size_column="final_expected_sample_size",
        ).charts
        assert len(charts) == 1

        charts = self.test.difference_plot(
            level_1=("1", "fin", "andr", "bananas_per_user_7d"),
            level_2=("2", "fin", "andr", "bananas_per_user_7d"),
            groupby="date",
            non_inferiority_margins=(0.01, "increase"),
            final_expected_sample_size_column="final_expected_sample_size",
        ).charts
        assert len(charts) == 1

    def test_multiple_difference_groupby(self):
        summary_df = self.test.summary()

        np.testing.assert_almost_equal(
            summary_df.query(
                'date == "2020-04-01" and group == "1" and country == "fin" '
                'and platform == "andr" and metric=="bananas_per_user_1d"'
            )[POINT_ESTIMATE].values[0],
            1.995389,
            5,
        )
        np.testing.assert_almost_equal(
            summary_df.query(
                'date == "2020-04-01" and group == "2" and country == "fin" '
                'and platform == "andr" and metric=="bananas_per_user_1d"'
            )[POINT_ESTIMATE].values[0],
            1.952098,
            5,
        )
        np.testing.assert_almost_equal(
            summary_df.query(
                'date == "2020-04-01" and group == "1" and country == "swe" '
                'and platform == "ios" and metric=="bananas_per_user_1d"'
            )[POINT_ESTIMATE].values[0],
            2.016416,
            5,
        )
        np.testing.assert_almost_equal(
            summary_df.query(
                'date == "2020-04-01" and group == "2" and country == "swe" '
                'and platform == "ios" and metric=="bananas_per_user_1d"'
            )[POINT_ESTIMATE].values[0],
            2.028478,
            5,
        )

        nims = {
            (
                pd.to_datetime("2020-04-01"),
                "bananas_per_user_1d",
                "fin",
                "andr",
            ): (None, None),
            (
                pd.to_datetime("2020-04-01"),
                "bananas_per_user_7d",
                "fin",
                "andr",
            ): (0.01, "increase"),
            (
                pd.to_datetime("2020-04-01"),
                "bananas_per_user_1d",
                "fin",
                "ios",
            ): (None, None),
            (
                pd.to_datetime("2020-04-01"),
                "bananas_per_user_7d",
                "fin",
                "ios",
            ): (0.01, "increase"),
            (
                pd.to_datetime("2020-04-01"),
                "bananas_per_user_1d",
                "swe",
                "andr",
            ): (None, None),
            (
                pd.to_datetime("2020-04-01"),
                "bananas_per_user_7d",
                "swe",
                "andr",
            ): (0.01, "increase"),
            (
                pd.to_datetime("2020-04-01"),
                "bananas_per_user_1d",
                "swe",
                "ios",
            ): (None, None),
            (
                pd.to_datetime("2020-04-01"),
                "bananas_per_user_7d",
                "swe",
                "ios",
            ): (0.01, "increase"),
            (
                pd.to_datetime("2020-04-02"),
                "bananas_per_user_1d",
                "fin",
                "andr",
            ): (None, None),
            (
                pd.to_datetime("2020-04-02"),
                "bananas_per_user_7d",
                "fin",
                "andr",
            ): (0.01, "increase"),
            (
                pd.to_datetime("2020-04-02"),
                "bananas_per_user_1d",
                "fin",
                "ios",
            ): (None, None),
            (
                pd.to_datetime("2020-04-02"),
                "bananas_per_user_7d",
                "fin",
                "ios",
            ): (0.01, "increase"),
            (
                pd.to_datetime("2020-04-02"),
                "bananas_per_user_1d",
                "swe",
                "andr",
            ): (None, None),
            (
                pd.to_datetime("2020-04-02"),
                "bananas_per_user_7d",
                "swe",
                "andr",
            ): (0.01, "increase"),
            (
                pd.to_datetime("2020-04-02"),
                "bananas_per_user_1d",
                "swe",
                "ios",
            ): (None, None),
            (
                pd.to_datetime("2020-04-02"),
                "bananas_per_user_7d",
                "swe",
                "ios",
            ): (0.01, "increase"),
        }

        difference_df = self.test.multiple_difference(
            level="1",
            groupby=["date", "metric", "country", "platform"],
            level_as_reference=True,
            final_expected_sample_size_column="final_expected_sample_size",
            non_inferiority_margins=nims,
        )
        assert len(difference_df) == (
            (self.data.group.unique().size - 1)
            * self.data.date.unique().size
            * self.data.country.unique().size
            * self.data.platform.unique().size
            * self.data.metric.unique().size
        )
        np.testing.assert_almost_equal(
            difference_df.query(
                'date == "2020-04-01" and country == "fin" and platform == "andr" and metric=="bananas_per_user_1d"'
            )[ADJUSTED_LOWER].values[0],
            -0.2016570,
            3,
        )
        np.testing.assert_almost_equal(
            difference_df.query(
                'date == "2020-04-01" and country == "fin" and platform == "andr" and metric=="bananas_per_user_1d"'
            )[ADJUSTED_UPPER].values[0],
            0.11507406,
            3,
        )
        np.testing.assert_almost_equal(
            difference_df.query(
                'date == "2020-04-01" and country == "swe" and platform == "ios" and metric=="bananas_per_user_1d"'
            )[ADJUSTED_LOWER].values[0],
            -0.1506963,
            3,
        )
        np.testing.assert_almost_equal(
            difference_df.query(
                'date == "2020-04-01" and country == "swe" and platform == "ios" and metric=="bananas_per_user_1d"'
            )[ADJUSTED_UPPER].values[0],
            0.17481994,
            3,
        )

        np.testing.assert_almost_equal(
            difference_df.query(
                'date == "2020-04-02" and country == "fin" and platform == "andr" and metric=="bananas_per_user_1d"'
            )[ADJUSTED_LOWER].values[0],
            -0.1063633,
            3,
        )
        np.testing.assert_almost_equal(
            difference_df.query(
                'date == "2020-04-02" and country == "fin" and platform == "andr" and metric=="bananas_per_user_1d"'
            )[ADJUSTED_UPPER].values[0],
            0.06637345,
            3,
        )
        np.testing.assert_almost_equal(
            difference_df.query(
                'date == "2020-04-02" and country == "swe" and platform == "ios" and metric=="bananas_per_user_1d"'
            )[ADJUSTED_LOWER].values[0],
            -0.0812409,
            3,
        )
        np.testing.assert_almost_equal(
            difference_df.query(
                'date == "2020-04-02" and country == "swe" and platform == "ios" and metric=="bananas_per_user_1d"'
            )[ADJUSTED_UPPER].values[0],
            0.09314668,
            3,
        )

        np.testing.assert_almost_equal(
            difference_df.query(
                'date == "2020-04-01" and country == "fin" and platform == "andr" and metric=="bananas_per_user_7d"'
            )[ADJUSTED_LOWER].values[0],
            -0.1932786,
            3,
        )
        np.isinf(
            difference_df.query(
                'date == "2020-04-01" and country == "fin" and platform == "andr" and metric=="bananas_per_user_7d"'
            )[ADJUSTED_UPPER].values[0]
        )
        np.testing.assert_almost_equal(
            difference_df.query(
                'date == "2020-04-01" and country == "swe" and platform == "ios" and metric=="bananas_per_user_7d"'
            )[ADJUSTED_LOWER].values[0],
            -0.1420855,
            3,
        )
        np.isinf(
            difference_df.query(
                'date == "2020-04-01" and country == "swe" and platform == "ios" and metric=="bananas_per_user_7d"'
            )[ADJUSTED_UPPER].values[0]
        )

        np.testing.assert_almost_equal(
            difference_df.query(
                'date == "2020-04-02" and country == "fin" and platform == "andr" and metric=="bananas_per_user_7d"'
            )[ADJUSTED_LOWER].values[0],
            -0.10027731,
            3,
        )
        np.isinf(
            difference_df.query(
                'date == "2020-04-02" and country == "fin" and platform == "andr" and metric=="bananas_per_user_7d"'
            )[ADJUSTED_UPPER].values[0]
        )
        np.testing.assert_almost_equal(
            difference_df.query(
                'date == "2020-04-02" and country == "swe" and platform == "ios" and metric=="bananas_per_user_7d"'
            )[ADJUSTED_LOWER].values[0],
            -0.07509674,
            3,
        )
        np.isinf(
            difference_df.query(
                'date == "2020-04-02" and country == "swe" and platform == "ios" and metric=="bananas_per_user_7d"'
            )[ADJUSTED_UPPER].values[0]
        )

        assert difference_df["p-value"].isnull().all()
        assert difference_df["adjusted p-value"].isnull().all()

        difference_df_2 = self.test.multiple_difference(
            level="1",
            groupby=["date", "metric", "country", "platform"],
            level_as_reference=True,
            final_expected_sample_size_column="final_expected_sample_size",
            non_inferiority_margins=True,
        )

        for column in difference_df.columns:
            assert (difference_df[column] == difference_df_2[column]).all() or (
                difference_df["p-value"].isnull() == difference_df_2["p-value"].isnull()
            ).all()


class TestSequentialOneSided(object):
    def setup(self):
        DATE = "date"
        COUNT = "count"
        SUM = "sum"
        SUM_OF_SQUARES = "sum_of_squares"
        GROUP = "group"

        self.data = pd.DataFrame(
            [
                {DATE: 1, GROUP: "1", COUNT: 1250, SUM: 2510.0, SUM_OF_SQUARES: 6304.0},
                {DATE: 1, GROUP: "2", COUNT: 1250, SUM: -2492.0, SUM_OF_SQUARES: 6163.0},
            ]
        ).assign(final_expected_sample_size=1e4)

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column=SUM,
            numerator_sum_squares_column=SUM_OF_SQUARES,
            denominator_column=COUNT,
            categorical_group_columns=GROUP,
            ordinal_group_column=DATE,
            interval_size=0.99,
        )

    def test_multiple_difference_groupby(self):
        difference_df = self.test.multiple_difference(
            level="1",
            groupby="date",
            level_as_reference=True,
            final_expected_sample_size_column="final_expected_sample_size",
            non_inferiority_margins=(0.01, "increase"),
        )

        assert len(difference_df) == ((self.data.group.unique().size - 1) * self.data.date.unique().size)
        assert difference_df["p-value"].isnull().all()
        assert difference_df["adjusted p-value"].isnull().all()
        assert np.isinf(difference_df[CI_UPPER].values[0])
        np.testing.assert_almost_equal(difference_df[ADJUSTED_LOWER].values[0], -4.129515314002298, 3)
        np.testing.assert_almost_equal(difference_df[DIFFERENCE].values[0], -4.001416, 3)


class TestSequentialTwoSided(object):
    def setup(self):
        DATE = "date"
        COUNT = "count"
        SUM = "sum"
        SUM_OF_SQUARES = "sum_of_squares"
        GROUP = "group"

        self.data = pd.DataFrame(
            [
                {"date": pd.to_datetime("2020-04-01"), GROUP: "1", COUNT: 1250, SUM: 2510.0, SUM_OF_SQUARES: 6304.0},
                {"date": pd.to_datetime("2020-04-01"), GROUP: "2", COUNT: 1250, SUM: 2492.0, SUM_OF_SQUARES: 6163.0},
            ]
        ).assign(final_expected_sample_size=1e4)

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column=SUM,
            numerator_sum_squares_column=SUM_OF_SQUARES,
            denominator_column=COUNT,
            categorical_group_columns=GROUP,
            ordinal_group_column=DATE,
            interval_size=0.99,
        )

    def test_multiple_difference_groupby(self):
        difference_df = self.test.multiple_difference(
            level="1",
            groupby="date",
            level_as_reference=True,
            final_expected_sample_size_column="final_expected_sample_size",
        )
        assert len(difference_df) == ((self.data.group.unique().size - 1) * self.data.date.unique().size)
        assert difference_df["p-value"].isnull().all()
        assert difference_df["adjusted p-value"].isnull().all()

        np.testing.assert_almost_equal(difference_df[ADJUSTED_UPPER].values[0], 0.121, 3)
        np.testing.assert_almost_equal(difference_df[ADJUSTED_LOWER].values[0], -0.151, 3)
        np.testing.assert_almost_equal(difference_df[DIFFERENCE].values[0], -0.0149, 3)


class TestSequentialOneSidedThreeGroups(object):
    def setup(self):
        DATE = "date"
        COUNT = "count"
        SUM = "sum"
        SUM_OF_SQUARES = "sum_of_squares"
        GROUP = "group"

        self.data = pd.DataFrame(
            [
                {DATE: 1, GROUP: "1", COUNT: 500, SUM: 1000, SUM_OF_SQUARES: 3000},
                {DATE: 1, GROUP: "2", COUNT: 1000, SUM: 1000, SUM_OF_SQUARES: 6000},
                {DATE: 1, GROUP: "3", COUNT: 1500, SUM: 1000, SUM_OF_SQUARES: 12000},
                {DATE: 2, GROUP: "1", COUNT: 1000, SUM: 2000, SUM_OF_SQUARES: 6000},
                {DATE: 2, GROUP: "2", COUNT: 2000, SUM: 2000, SUM_OF_SQUARES: 12000},
                {DATE: 2, GROUP: "3", COUNT: 3000, SUM: 2000, SUM_OF_SQUARES: 24000},
            ]
        ).assign(final_expected_sample_size=1e4)

        self.test = spotify_confidence.ZTest(
            self.data,
            numerator_column=SUM,
            numerator_sum_squares_column=SUM_OF_SQUARES,
            denominator_column=COUNT,
            categorical_group_columns=[DATE],
            ordinal_group_column=DATE,
            treatment_column=GROUP,
            interval_size=0.95,
        )

    def test_multiple_difference_groupby(self):
        difference_df = self.test.multiple_difference(
            level="1",
            groupby="date",
            level_as_reference=True,
            final_expected_sample_size_column="final_expected_sample_size",
        )

        test_df = difference_df.reset_index().sort_values(["level_2", "date"])
        np.testing.assert_almost_equal(test_df[ADJUSTED_LOWER].values, [-1.327445, -1.209182, -1.646339, -1.531049], 3)
        np.testing.assert_almost_equal(
            test_df[ADJUSTED_UPPER].values, [-0.6725549, -0.7908185, -1.0203281, -1.1356181], 3
        )


class TestNimsWithNaN(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "count": {
                    0: 252934,
                    1: 253656,
                    2: 252328,
                    3: 464640,
                    4: 465726,
                    5: 465194,
                    6: 463493,
                    7: 464487,
                    8: 464059,
                },
                "sum": {
                    0: 89984.0,
                    1: 89992.0,
                    2: 89108.0,
                    3: 5815.0,
                    4: 5867.0,
                    5: 5896.0,
                    6: 13928.0,
                    7: 13688.0,
                    8: 13505.0,
                },
                "sum_of_squares": {
                    0: 89984.0,
                    1: 89992.0,
                    2: 89108.0,
                    3: 5815.0,
                    4: 5867.0,
                    5: 5896.0,
                    6: 13928.0,
                    7: 13688.0,
                    8: 13505.0,
                },
                "exposure_experiment_group_id": {
                    0: "Control",
                    1: "Treatment1",
                    2: "Treatment2",
                    3: "Control",
                    4: "Treatment1",
                    5: "Treatment2",
                    6: "Control",
                    7: "Treatment1",
                    8: "Treatment2",
                },
                "non_inferiority_margin": {
                    0: np.nan,
                    1: np.nan,
                    2: np.nan,
                    3: np.nan,
                    4: np.nan,
                    5: np.nan,
                    6: 5.0,
                    7: 5.0,
                    8: 5.0,
                },
                "preferred_direction": {
                    0: np.nan,
                    1: np.nan,
                    2: np.nan,
                    3: "DECREASE",
                    4: "DECREASE",
                    5: "DECREASE",
                    6: "INCREASE",
                    7: "INCREASE",
                    8: "INCREASE",
                },
                "metric": {0: "m1", 1: "m1", 2: "m1", 3: "m2", 4: "m2", 5: "m2", 6: "m3", 7: "m3", 8: "m3"},
            }
        )

    def test_nims_with_nans(self):
        ztest = spotify_confidence.ZTest(
            data_frame=self.data,
            numerator_column="sum",
            numerator_sum_squares_column="sum_of_squares",
            denominator_column="count",
            categorical_group_columns=["metric", "exposure_experiment_group_id"],
            interval_size=0.99,
        )

        diff_df = ztest.multiple_difference(
            level="Control", level_as_reference=True, absolute=False, groupby="metric", non_inferiority_margins=True
        )

        assert len(diff_df) == 6



import spotify_confidence
import pandas as pd

OUTPUT_DIR = "./tests/outputs/precision/"

spotify_confidence.options.set_option("randomization_seed", 1)


def test_bayesian_precision_plots():
    data = pd.DataFrame(
        {
            "variation_name": [
                "test",
                "control",
            ],
            "success": [
                50000000,
                50004000,
            ],
            "total": [
                100900000,
                100900000,
            ],
        }
    )

    test = spotify_confidence.BetaBinomial(
        data, numerator_column="success", denominator_column="total", categorical_group_columns=["variation_name"]
    )

    test.summary_plot().charts[0].save(OUTPUT_DIR + "bayesian_1.html", "html")

    data = pd.DataFrame(
        {
            "variation_name": ["test", "control", "test2"],
            "success": [
                50000000,
                50004000,
                100500000,
            ],
            "total": [
                100900000,
                100900000,
                150200000,
            ],
        }
    )

    test = spotify_confidence.BetaBinomial(
        data, numerator_column="success", denominator_column="total", categorical_group_columns=["variation_name"]
    )

    test.summary_plot().charts[0].save(OUTPUT_DIR + "bayesian_2.html", "html")

    test.difference_plot("control", "test").charts[0].save(OUTPUT_DIR + "bayesian_3.html", "html")

    test.difference_plot("control", "test2").charts[0].save(OUTPUT_DIR + "bayesian_4.html", "html")

    test.multiple_difference_plot("test2").charts[0].save(OUTPUT_DIR + "bayesian_5.html", "html")


def test_frequentist_precision_plots():
    data = pd.DataFrame(
        {
            "variation_name": ["test", "control", "test2"],
            "success": [
                50000000,
                50004000,
                100500000,
            ],
            "total": [
                100900000,
                100900000,
                150200000,
            ],
        }
    )

    test = spotify_confidence.ChiSquared(
        data, numerator_column="success", denominator_column="total", categorical_group_columns=["variation_name"]
    )

    test.summary_plot().charts[0].save(OUTPUT_DIR + "frequentist_1.html", "html")

    test.difference_plot("control", "test").charts[0].save(OUTPUT_DIR + "frequentist_2.html", "html")

    test.difference_plot("control", "test2").charts[0].save(OUTPUT_DIR + "frequentist_3.html", "html")

    test = spotify_confidence.ChiSquared(
        data.head(2),
        numerator_column="success",
        denominator_column="total",
        categorical_group_columns=["variation_name"],
    )
    test.summary_plot().charts[0].save(OUTPUT_DIR + "frequentist_4.html", "html")


def test_ordinal_precision_plots():
    data = pd.DataFrame(
        {
            "variation_name": [
                "test",
                "control",
                "test2",
                "test",
                "control",
                "test2",
                "test",
                "control",
                "test2",
                "test",
                "control",
                "test2",
                "test",
                "control",
                "test2",
            ],
            "success": [
                500,
                8,
                100,
                510,
                8,
                100,
                520,
                9,
                104,
                520,
                7,
                100,
                530,
                8,
                103,
            ],
            "total": [
                1010,
                22,
                150,
                1000,
                20,
                153,
                1030,
                23,
                154,
                1000,
                20,
                150,
                1040,
                21,
                155,
            ],
            "days_since_reg": [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5],
        }
    )

    test = spotify_confidence.BetaBinomial(
        data,
        numerator_column="success",
        denominator_column="total",
        categorical_group_columns="variation_name",
        ordinal_group_column="days_since_reg",
    )

    test.summary_plot().charts[0].save(OUTPUT_DIR + "ordinal_1.html", "html")
    test.summary_plot("variation_name").charts[0].save(OUTPUT_DIR + "ordinal_2.html", "html")

    data = pd.DataFrame(
        {
            "variation_name": [
                "test",
                "control",
                "test2",
                "test",
                "control",
                "test2",
                "test",
                "control",
                "test2",
                "test",
                "control",
                "test2",
                "test",
                "control",
                "test2",
            ],
            "success": [
                500000,
                800,
                10000,
                500200,
                800,
                10000,
                500100,
                900,
                10400,
                500100,
                700,
                10000,
                500200,
                800,
                10300,
            ],
            "total": [
                1000010,
                2002,
                15000,
                1000000,
                2000,
                15300,
                1000030,
                2003,
                15400,
                1000000,
                2000,
                15000,
                1000000,
                2001,
                15500,
            ],
            "days_since_reg": [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5],
        }
    )

    test = spotify_confidence.BetaBinomial(
        data,
        numerator_column="success",
        denominator_column="total",
        categorical_group_columns="variation_name",
        ordinal_group_column="days_since_reg",
    )
    test.summary_plot().charts[0].save(OUTPUT_DIR + "ordinal_3.html", "html")
    test.summary_plot("variation_name").charts[0].save(OUTPUT_DIR + "ordinal_4.html", "html")



# Copyright 2017-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, Iterable, Tuple, Dict, List

from pandas import DataFrame

from spotify_confidence.analysis.frequentist.confidence_computers.confidence_computer import ConfidenceComputer
from .chartify_grapher import ChartifyGrapher
from ..abstract_base_classes.confidence_abc import ConfidenceABC
from ..abstract_base_classes.confidence_computer_abc import ConfidenceComputerABC
from ..abstract_base_classes.confidence_grapher_abc import ConfidenceGrapherABC
from ..confidence_utils import (
    validate_categorical_columns,
    listify,
    get_all_categorical_group_columns,
    get_all_group_columns,
)
from ..constants import BONFERRONI, NIM_TYPE, METHODS
from ..frequentist.sample_ratio_test import sample_ratio_test
from ...chartgrid import ChartGrid


class Experiment(ConfidenceABC):
    def __init__(
        self,
        data_frame: DataFrame,
        numerator_column: str,
        numerator_sum_squares_column: Union[str, None],
        denominator_column: str,
        categorical_group_columns: Union[str, Iterable],
        ordinal_group_column: Union[str, None] = None,
        interval_size: float = 0.95,
        correction_method: str = BONFERRONI,
        confidence_computer: ConfidenceComputerABC = None,
        confidence_grapher: ConfidenceGrapherABC = None,
        method_column: str = None,
        bootstrap_samples_column: str = None,
        metric_column=None,
        treatment_column=None,
        power: float = 0.8,
        feature_column: str = None,
        feature_sum_squares_column: str = None,
        feature_cross_sum_column: str = None,
    ):
        validate_categorical_columns(categorical_group_columns)
        self._df = data_frame
        self._numerator = numerator_column
        self._numerator_sumsq = numerator_sum_squares_column
        self._denominator = denominator_column
        self._categorical_group_columns = get_all_categorical_group_columns(
            categorical_group_columns, metric_column, treatment_column
        )
        self._ordinal_group_column = ordinal_group_column
        self._metric_column = metric_column
        self._treatment_column = treatment_column
        self._all_group_columns = get_all_group_columns(self._categorical_group_columns, self._ordinal_group_column)
        if method_column is None:
            raise ValueError("method column cannot be None")
        if not all(self._df[method_column].map(lambda m: m in METHODS)):
            raise ValueError(f"Values of method column must be in {METHODS}")

        if confidence_computer is not None:
            self._confidence_computer = confidence_computer
        else:
            self._confidence_computer = ConfidenceComputer(
                data_frame=data_frame,
                numerator_column=numerator_column,
                numerator_sum_squares_column=numerator_sum_squares_column,
                denominator_column=denominator_column,
                categorical_group_columns=listify(categorical_group_columns),
                ordinal_group_column=ordinal_group_column,
                interval_size=interval_size,
                correction_method=correction_method.lower(),
                method_column=method_column,
                bootstrap_samples_column=bootstrap_samples_column,
                metric_column=metric_column,
                treatment_column=treatment_column,
                power=power,
                feature_column=feature_column,
                feature_sum_squares_column=feature_sum_squares_column,
                feature_cross_sum_column=feature_cross_sum_column,
            )

        self._confidence_grapher = (
            confidence_grapher
            if confidence_grapher is not None
            else ChartifyGrapher(
                data_frame=self._df,
                numerator_column=self._numerator,
                denominator_column=self._denominator,
                categorical_group_columns=self._categorical_group_columns,
                ordinal_group_column=self._ordinal_group_column,
            )
        )

    def summary(self, verbose: bool = False) -> DataFrame:
        return self._confidence_computer.compute_summary(verbose)

    def difference(
        self,
        level_1: Union[str, Tuple],
        level_2: Union[str, Tuple],
        absolute: bool = True,
        groupby: Union[str, Iterable] = None,
        non_inferiority_margins: NIM_TYPE = None,
        final_expected_sample_size_column: str = None,
        verbose: bool = False,
        minimum_detectable_effects_column: str = None,
    ) -> DataFrame:
        self._validate_sequential(final_expected_sample_size_column, groupby)

        return self._confidence_computer.compute_difference(
            level_1,
            level_2,
            absolute,
            groupby,
            non_inferiority_margins,
            final_expected_sample_size_column,
            verbose,
            minimum_detectable_effects_column,
        )

    def differences(
        self,
        levels: Union[Tuple, List[Tuple]],
        absolute: bool = True,
        groupby: Union[str, Iterable] = None,
        non_inferiority_margins: NIM_TYPE = None,
        final_expected_sample_size_column: str = None,
        verbose: bool = False,
        minimum_detectable_effects_column: str = None,
    ) -> DataFrame:
        self._validate_sequential(final_expected_sample_size_column, groupby)
        return self._confidence_computer.compute_differences(
            levels,
            absolute,
            groupby,
            non_inferiority_margins,
            final_expected_sample_size_column,
            verbose,
            minimum_detectable_effects_column,
        )

    def multiple_difference(
        self,
        level: Union[str, Tuple],
        absolute: bool = True,
        groupby: Union[str, Iterable] = None,
        level_as_reference: bool = None,
        non_inferiority_margins: NIM_TYPE = None,
        final_expected_sample_size_column: str = None,
        verbose: bool = False,
        minimum_detectable_effects_column: str = None,
    ) -> DataFrame:
        self._validate_sequential(final_expected_sample_size_column, groupby)

        return self._confidence_computer.compute_multiple_difference(
            level,
            absolute,
            groupby,
            level_as_reference,
            non_inferiority_margins,
            final_expected_sample_size_column,
            verbose,
            minimum_detectable_effects_column,
        )

    def summary_plot(self, groupby: Union[str, Iterable] = None) -> ChartGrid:
        summary_df = self.summary()
        graph = self._confidence_grapher.plot_summary(summary_df, groupby)
        return graph

    def difference_plot(
        self,
        level_1: Union[str, Tuple],
        level_2: Union[str, Tuple],
        absolute: bool = True,
        groupby: Union[str, Iterable] = None,
        non_inferiority_margins: NIM_TYPE = None,
        use_adjusted_intervals: bool = False,
        final_expected_sample_size_column: str = None,
        split_plot_by_groups: bool = False,
    ) -> ChartGrid:
        difference_df = self.difference(
            level_1=level_1,
            level_2=level_2,
            absolute=absolute,
            groupby=groupby,
            non_inferiority_margins=non_inferiority_margins,
            final_expected_sample_size_column=final_expected_sample_size_column,
            verbose=True,
        )
        chartgrid = self._confidence_grapher.plot_difference(
            difference_df, absolute, groupby, non_inferiority_margins, use_adjusted_intervals, split_plot_by_groups
        )
        return chartgrid

    def differences_plot(
        self,
        levels: List[Tuple],
        absolute: bool = True,
        groupby: Union[str, Iterable] = None,
        non_inferiority_margins: NIM_TYPE = None,
        use_adjusted_intervals: bool = False,
        final_expected_sample_size_column: str = None,
        split_plot_by_groups: bool = False,
    ) -> ChartGrid:
        difference_df = self.differences(
            levels, absolute, groupby, non_inferiority_margins, final_expected_sample_size_column, verbose=True
        )
        chartgrid = self._confidence_grapher.plot_differences(
            difference_df, absolute, groupby, non_inferiority_margins, use_adjusted_intervals, split_plot_by_groups
        )
        return chartgrid

    def multiple_difference_plot(
        self,
        level: Union[str, Tuple],
        absolute: bool = True,
        groupby: Union[str, Iterable] = None,
        level_as_reference: bool = None,
        non_inferiority_margins: NIM_TYPE = None,
        use_adjusted_intervals: bool = False,
        final_expected_sample_size_column: str = None,
        split_plot_by_groups: bool = False,
    ) -> ChartGrid:
        difference_df = self.multiple_difference(
            level=level,
            absolute=absolute,
            groupby=groupby,
            level_as_reference=level_as_reference,
            non_inferiority_margins=non_inferiority_margins,
            final_expected_sample_size_column=final_expected_sample_size_column,
            verbose=True,
        )
        chartgrid = self._confidence_grapher.plot_multiple_difference(
            difference_df,
            absolute,
            groupby,
            level_as_reference,
            non_inferiority_margins,
            use_adjusted_intervals,
            split_plot_by_groups,
        )
        return chartgrid

    def sample_ratio_test(self, expected_proportions: Dict) -> Tuple[float, DataFrame]:
        return sample_ratio_test(
            self._df,
            all_group_columns=self._all_group_columns,
            denominator=self._denominator,
            expected_proportions=expected_proportions,
        )

    def achieved_power(self, level_1, level_2, mde, alpha, groupby=None) -> DataFrame:
        """Calculated the achieved power of test of differences between
        level 1 and level 2 given a targeted MDE.

        Args:
            level_1 (str, tuple of str): Name of first level.
            level_2 (str, tuple of str): Name of second level.
            mde (float): Absolute minimal detectable effect size.
            alpha (float): Type I error rate, cutoff value for determining
                statistical significance.
            groupby (str): Name of column.
                If specified, will return the difference for each level
                of the grouped dimension.

        Returns:
            Pandas DataFrame with the following columns:
            - level_1: Name of level 1.
            - level_2: Name of level 2.
            - power: 1 - B, where B is the likelihood of a Type II (false
                negative) error.

        """
        return self._confidence_computer.achieved_power(level_1, level_2, mde, alpha, groupby)

    def _validate_sequential(self, final_expected_sample_size: float, groupby: Union[str, Iterable]):
        if final_expected_sample_size is not None:
            if self._ordinal_group_column not in listify(groupby):
                raise ValueError(
                    f"{self._ordinal_group_column} must be in groupby argument to use "
                    f"sequential testing with final_expected_sample_size"
                )



from numpy import isnan
from pandas import DataFrame

from spotify_confidence.analysis.confidence_utils import listify
from spotify_confidence.analysis.constants import (
    NIM_TYPE,
    NIM_COLUMN_DEFAULT,
    PREFERRED_DIRECTION_COLUMN_DEFAULT,
    ORIGINAL_POINT_ESTIMATE,
    TWO_SIDED,
    INCREASE_PREFFERED,
    DECREASE_PREFFERED,
    NIM,
    PREFERENCE,
    NULL_HYPOTHESIS,
    ALTERNATIVE_HYPOTHESIS,
)


def add_nim_input_columns_from_tuple_or_dict(df, nims: NIM_TYPE, mde_column: str) -> DataFrame:
    if type(nims) is tuple:
        return df.assign(**{NIM_COLUMN_DEFAULT: nims[0]}).assign(**{PREFERRED_DIRECTION_COLUMN_DEFAULT: nims[1]})
    elif type(nims) is dict:
        nim_values = {key: value[0] for key, value in nims.items()}
        nim_preferences = {key: value[1] for key, value in nims.items()}
        return df.assign(**{NIM_COLUMN_DEFAULT: lambda df: df.index.to_series().map(nim_values)}).assign(
            **{PREFERRED_DIRECTION_COLUMN_DEFAULT: lambda df: df.index.to_series().map(nim_preferences)}
        )
    elif nims is None or not nims:
        return df.assign(**{NIM_COLUMN_DEFAULT: None}).assign(
            **{
                PREFERRED_DIRECTION_COLUMN_DEFAULT: None
                if PREFERRED_DIRECTION_COLUMN_DEFAULT not in df or mde_column is None
                else df[PREFERRED_DIRECTION_COLUMN_DEFAULT]
            }
        )
    else:
        return df


def add_nims_and_mdes(
    df: DataFrame,
    mde_column: str,
    nim_column: str,
    preferred_direction_column: str,
) -> DataFrame:
    def _set_nims_and_mdes(grp: DataFrame) -> DataFrame:
        nim = grp[nim_column].astype(float)
        input_preference = grp[preferred_direction_column].values[0]
        mde = None if mde_column is None else grp[mde_column]

        nim_is_na = nim.isna().all()
        mde_is_na = True if mde is None else mde.isna().all()
        if input_preference is None or (type(input_preference) is float and isnan(input_preference)):
            signed_nim = 0.0 if nim_is_na else nim * grp[ORIGINAL_POINT_ESTIMATE]
            preference = TWO_SIDED
            signed_mde = None if mde_is_na else mde * grp[ORIGINAL_POINT_ESTIMATE]
        elif input_preference.lower() == INCREASE_PREFFERED:
            signed_nim = 0.0 if nim_is_na else -nim * grp[ORIGINAL_POINT_ESTIMATE]
            preference = "larger"
            signed_mde = None if mde_is_na else mde * grp[ORIGINAL_POINT_ESTIMATE]
        elif input_preference.lower() == DECREASE_PREFFERED:
            signed_nim = 0.0 if nim_is_na else nim * grp[ORIGINAL_POINT_ESTIMATE]
            preference = "smaller"
            signed_mde = None if mde_is_na else -mde * grp[ORIGINAL_POINT_ESTIMATE]
        else:
            raise ValueError(f"{input_preference.lower()} not in " f"{[INCREASE_PREFFERED, DECREASE_PREFFERED]}")

        return (
            grp.assign(**{NIM: nim})
            .assign(**{PREFERENCE: preference})
            .assign(**{NULL_HYPOTHESIS: signed_nim})
            .assign(**{ALTERNATIVE_HYPOTHESIS: signed_mde if nim_is_na else 0.0})
        )

    index_names = [name for name in df.index.names if name is not None]
    return (
        df.groupby(
            [nim_column, preferred_direction_column] + listify(mde_column),
            dropna=False,
            as_index=False,
            sort=False,
            group_keys=True,
        )
        .apply(_set_nims_and_mdes)
        .pipe(lambda df: df.reset_index(index_names))
        .reset_index(drop=True)
        .pipe(lambda df: df if index_names == [] else df.set_index(index_names))
    )
